/* 
 * wsdlpull - A C++ parser  for WSDL  (Web services description language)
 * Copyright (C) 2003 Vivek Krishna
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *
 */

#include <gsoap.h>
#include <sstream>
#include <map>
extern int *isDefnGenerated;
map < string, string > prefixes;
void
gsoap_CopyFile2Stream (const char *file_name, ostream * os)
{
  ifstream file (file_name);
  for (string s; getline (file, s); (*os) << s << endl);
  (*os) << endl;
  file.close ();
  (*os).flush ();
}

string
xsd2CmapTable (int typeId)
{
  switch (typeId)

    {
    case XSDType::STRING:
    case XSDType::ANYURI:
    case XSDType::NMTOKEN:
    case XSDType::NMTOKENS:
    case XSDType::NCNAME:
    case XSDType::TOKEN:
      return "xsd__string";
    case XSDType::INTEGER:
      return "xsd__integer_";
    case XSDType::INT:
      return "xsd__int";
    case XSDType::BYTE:
      return "xsd__byte";
    case XSDType::POSINT:
      return "xsd__positiveInteger";
    case XSDType::FLOAT:
      return "xsd__float";
    case XSDType::LONG:
      return "xsd__long";
    case XSDType::DECIMAL:
      return "xsd__decimal";
    case XSDType::BOOLEAN:
      return "xsd__boolean";
    case XSDType::ANY:
    case XSDType::ANYTYPE:
      return "XML";
    case XSDType::DOUBLE:
      return "xsd__double";
    case XSDType::BASE64BIN:
      return "struct xsd__base64Binary ";
    case XSDType::HEXBIN:
      return "struct xsd__hexBinary ";
    case XSDType::DATE:
      return "xsd__date";
    case XSDType::DATETIME:
      return "xsd__dateTime";
    case XSDType::TIME:
      return "xsd__time";
    default:
      return "";
    }
}


//this function returns,or generates if necessary a unique prefix for each namespace
string
gsoap_getNamespacePrefix (string ns)
{
  static int nsCount = 0;
  map < string, string >::iterator it = prefixes.find (ns);
  if (it == prefixes.end ())

    {
      string pre;
      ostringstream ostr;
      ostr << "ns" << nsCount;
      prefixes[ns] = pre = ostr.str ();
      nsCount++;
      return pre;
    }

  else

    {
      return prefixes[ns];
    }
}

bool
gsoap_createDirective (const string & nsPrefix, const string & wsdlPart,
		       const string & param, const string & value,
		       ostream * os)
{
  (*os) << "//gsoap" << dbsp << nsPrefix << dbsp << wsdlPart << dbsp <<
    param << ":" << dbsp << value << endl;
  return true;
}

//need to codify XSD list type elements like arrays for gsoap
bool
gsoap_isArray (ComplexType * ct)
{
  if (ct->isArray ())
    return true;
  if (ct->getNumChildren () == 1)

    {
      Element & e = ct->getElement (0);
      if (e.max () == UNBOUNDED)
	return true;
    }
  return false;
}

//This function generates a qualified type prefix
//like xsd__int  , struct ns__ResultArray , enum ns__Status or ns__Result
string
gsoap_generateQualifiedType (int typeId, const SchemaParser * sParser)
{
  const XSDType *pType = sParser->getType (typeId);
  string sPrefix;
  if (pType)			//take the prefix from the pType,it could be an ext Ref from another schema
    sPrefix = gsoap_getNamespacePrefix ((pType->getQname ()).getNamespace ());

  else
    sPrefix = gsoap_getNamespacePrefix (sParser->getTargetNamespace ());
  if (pType)

    {
      if (pType->isSimple ())
	return "enum  " + sPrefix + "__" + pType->getName ();

      else

	{
	  if (gsoap_isArray ((ComplexType *) pType))
	    {
	      if (((ComplexType *) pType)->isArray ())
		return "struct " + pType->getName ();

	      else
		return "struct " + sPrefix + "__" + pType->getName ();
	    }
	  else
	    {
	      return sPrefix + "__" + pType->getName ();
	    }
	}
    }

  else
    return xsd2CmapTable (typeId);
}



void
gsoap_generateParticleAndAttrs (ComplexType * cType,
				const SchemaParser * sParser, ostream * os)
{
  int i = 0;
  for (i = 0; i < cType->getNumAttributes (); i++)

    {
      Attribute at = cType->getAttribute (i);
      (*os) << "@" << gsoap_generateQualifiedType (at.getType (),
						   sParser) << dbsp <<
	at.getName () << ";" << endl;
    }
  for (i = 0; i < cType->getNumChildren (); i++)

    {
      Element e = cType->getElement (i);
      //if the child is anonymous generate it inline
      const XSDType *pChild = sParser->getType (e.getType ());
      if (pChild && !pChild->isSimple () && pChild->isAnonymous ())
	{
	  (*os) << endl;
	  (*os) << "struct {" << endl;
	  gsoap_generateParticleAndAttrs ((ComplexType *) pChild, sParser,
					  os);
	  (*os) << "}" << e.getName () << ";" << endl;
	}
      else
	{
	  (*os) << gsoap_generateQualifiedType (e.getType (),
						sParser) << dbsp << e.
	    getName () << ";" << endl;
	}
    }

}

//this is the main function which creates the structs for 
//the types defined in the schema
bool
gsoap_createStruct (const XSDType * pType, const SchemaParser * sParser,
		    ostream * os, string nsPrefix)
{
  string sPrefix =		//use the schema's target namespace if none is  provided
    (nsPrefix.empty ())? gsoap_getNamespacePrefix (sParser->
						   getTargetNamespace ())
    : nsPrefix;

//if the type does not belong to this namespace but is just refernced then
//dont generate it
  if ((pType->getQname ()).getNamespace () != sParser->getTargetNamespace ())
    return true;

//if its a complex type ensure that all types it references are generated
  if (!pType->isSimple ())
    {
      ComplexType *ct = (ComplexType *) pType;

      //generate the base type if present
      if (ct->getBaseType () != XSDType::ANYTYPE
	  && !isDefnGenerated[ct->getBaseType () - (XSDType::ANYURI + 1)])
	gsoap_createStruct (sParser->getType (ct->getBaseType ()), sParser,
			    os);

      //generate structs for the types referenced by its children
      for (int j = 0; j < ct->getNumChildren (); j++)
	if (!sParser->isBasicType (ct->getChildType (j))
	    && !isDefnGenerated[ct->getChildType (j) - (XSDType::ANYURI + 1)])
	  {
	    const XSDType *pChild = sParser->getType (ct->getChildType (j));
	    if (pChild->isAnonymous ())
	      continue;
	    if (!gsoap_createStruct (pChild, sParser, os))
	      return false;
	    isDefnGenerated[ct->getChildType (j) -
			    (XSDType::ANYURI + 1)] = true;
	  }
    }
  if (pType->isSimple ())

    {
      SimpleType *st = (SimpleType *) pType;
      list < string > *lStrEnums;

      if (st->getFacetValue (SimpleType::ENUM, (void **) &lStrEnums))

	{
	  list < string >::iterator pEnums;
	  bool first = true;
	  (*os) << "enum" << dbsp;
	  (*os) << sPrefix << "__" << st->getName () << "{" << dbsp;
	  for (pEnums = (*lStrEnums).begin ();
	       pEnums != (*lStrEnums).end (); pEnums++)

	    {
	      if (!first)
		(*os) << ",";
	      (*os) << *pEnums << dbsp;
	      first = false;
	    }
	  (*os) << "}" << ";" << endl;
	}

      else

	{
	  (*os) << "typedef" << dbsp << xsd2CmapTable (st->
						       getBaseType ()) <<
	    dbsp;
	  (*os) << sPrefix << "__" << st->getName () << ";" << endl;
	}
    }

  else
    {
      ComplexType *ct = (ComplexType *) pType;
      int i = 0;
      if (ct->getContentModel () == ct->SIMPLE)
	return false;		// no support for this in gsoap
      if (gsoap_isArray (ct))
	{
	  if (ct->isArray ())
	    (*os) << "struct" << dbsp << ct->getName () << "{" << endl;

	  else			//gsoap..this is an array but must be serialised as a list arrays are generated as structs
	    (*os) << "struct" << dbsp << sPrefix << "__" << ct->
	      getName () << "{" << endl;
	  (*os) << "int __size;" << endl;
	  Element & e = ct->getElement (0);
	  const XSDType *pType = sParser->getType (e.getType ());
	  if (pType && pType->isAnonymous () && !pType->isSimple ())
	    {
	      (*os) << "struct {" << endl;
	      gsoap_generateParticleAndAttrs ((ComplexType *) pType, sParser,
					      os);
	      (*os) << "} * " << e.getName () << ";" << endl;
	    }
	  else
	    {

	      (*os) << gsoap_generateQualifiedType (e.getType (),
						    sParser) << "*" << dbsp <<
		e.getName () << "_;" << endl;
	    }
	}

      else

	{
	  (*os) << "class" << dbsp << sPrefix << "__" << ct->getName ();
	  if (ct->getBaseType () != XSDType::ANYTYPE)
	    {
	      (*os) << ": public " << gsoap_generateQualifiedType (ct->
								   getBaseType
								   (),
								   sParser);
	    }
	  (*os) << "{" << endl;
	  gsoap_generateParticleAndAttrs (ct, sParser, os);

	}
      (*os) << "};" << endl;
    }
  isDefnGenerated[pType->getTypeId () - (XSDType::ANYURI + 1)] = true;
  return true;
}

void
gsoap_generate_HeaderStruct (int headerTypeId, const SchemaParser * sParser,
			     ostream * os)
{
  XSDType *pType = (XSDType *) sParser->getType (headerTypeId);
  string name = pType->getName ();
  pType->setName ("Header");	//crude but works !!
  gsoap_createStruct (pType, sParser, os, "SOAP_ENV");
  pType->setName (name);
}

string
gsoap_createRemoteMethod (string Ns, string methodName, ostream * os)
{
  (*os) << "int " << Ns << "__" << methodName << "(" << dbsp;
  return Ns+"__"+methodName;
}

bool
gsoap_addInParam (string typeName, string paramName, ostream * os)
{
  if (paramName == "return")
    paramName = "return_";
  (*os) << dbsp << typeName << dbsp << "*" << paramName << "_," << dbsp;
  return true;
}

bool
gsoap_addOutParam (string typeName, string paramName, ostream * os)
{
  if (paramName == "return")
    paramName = "return_";
  (*os) << dbsp << typeName << "*" << dbsp << paramName << "_" << dbsp;
  return true;
}

bool
gsoap_finishRemoteMethod (ostream * os)
{
  (*os) << ");" << endl;
  return true;
}
