/* 
 * wsdlpull - A C++ parser  for WSDL  (Web services description language)
 * Copyright (C) 2003 Vivek Krishna
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *
 */

#include <schemaparser/ComplexType.h>
ComplexType::ComplexType(string ns)
  :nsUri(ns)
{
  typeId = 0;                                   //this is a unique id
  num_elems = 0;
  bArray = false;
  anonymous = false;
  num_atts = 0;
  contentModel = COMPLEX;
  mixed = false;
  currentGroup = prevGroup = 0;
  topLevelGroup = ALL;
  contentTypeId = 0;
  baseType.typeId = XSDType::ANYTYPE;
  baseType.derivation= Extension;
  baseType.type=0;
  
  group[0].type = group[1].type = group[2].type = group[3].type =
    group[4].type = ALL;
  fwdElemRef = fwdAttrRef = false;
}


void ComplexType::addElement(const string & nam, 
                             int type_id, 
                             int minOccurs,
                             int maxOccurs, 
                             bool qualified, 
                             string def,
                             string fixed, 
                             bool append)
{
  if (type_id == 0)
    error("ComplexType:addElement() :No type given for element ");
  if (topLevelGroup == ALL && maxOccurs > 1 && bArray == false)
    error("Inside an <all> group elements can occur only once");
  Element e(nam, type_id, minOccurs, maxOccurs, qualified, def, fixed);
  e.setGroupId(currentGroup);
  Element *el=(Element *)getElement(nam);
  
  if(el){
    //element is probably a base type's child
    *el=e;
    
  }else{
    elem_list.push_back(e);
    num_elems++;
  }
  
}


void ComplexType::addElementRef(Qname elem, int min, int max)
{
  addElement(elem.getLocalName(), XSDType::ANYTYPE, min, max);
  fwdElemRef = true;
}


void ComplexType::matchElementRef(const string & nam, Element & e)
{
  if (fwdElemRef)
    {
      Element *el = (Element*) getElement(nam);
      if (el){
        //these values are specific to the occurence of the element inside another XML type
        //so we shouldnt delete them
        int min=el->min(); 
        int max=el->max();
        int gId=el->getGroupId();
        *el = e;
        el->min(min);
        el->max(max);
        el->setGroupId(gId);
      }
    }
}


void ComplexType::addAttribute(const string & nam, 
                               int type_id, 
                               bool qualified,
                               string defaultVal, 
                               string fixedVal,
                               bool use, 
                               bool append)
{
  if (type_id == 0)
    error("ComplexType:addAttribute() :No type given for attribute");
  Attribute a(nam, type_id, qualified, defaultVal, fixedVal, use);
 
  Attribute *at= (Attribute *)getAttribute(nam);
  if(at){
    *at=a;
  }else{

    att_list.push_back(a);
    num_atts++;
  }
  
}


void ComplexType::addAttributeRef(Qname attr, 
                                  bool qualified, 
                                  bool use)
{
  addAttribute(attr.getLocalName(), XSDType::ANYTYPE, qualified, "", "",
               use);
  fwdAttrRef = true;
}


// 'name' hides member.  Changed to 'nam'.  -KEC
void ComplexType::matchAttributeRef(const string & nam, 
                                    Attribute & a)
{
  if (fwdAttrRef)
    {
      Attribute *at = (Attribute*)getAttribute(nam);
      if (at)
        *at = a;
    }
}


//once a compositor is set  ,all subsequent elements added are restrained by the last added compositor(or the active one)
void ComplexType::setCompositor(Compositor type, 
                                bool open, 
                                int min, 
                                int max)
{

  //type is one of SEQ,CHOICE,ALL.open is true if its the block begin
  if (open)
    {
      prevGroup = currentGroup++;
      if (prevGroup == 0)
        topLevelGroup = type;

      else if (topLevelGroup == this->ALL)
        error("This cannot occur inside a top level <all> compositor");
      if (type == this->ALL && prevGroup != 0)
        error("<all> can occur only at the top level");
      if (type == this->ALL && (min != 1 || max != 1))
        error("<all> can have min/max of only 1 ");
      group[currentGroup].type = type;
      group[currentGroup].min = min;
      group[currentGroup].max = max;
    }

  else
    {
      currentGroup = prevGroup;
    }
}


Element & ComplexType::getElement(int id)
{
  if (id < 0 || id >= num_elems)
    error("tried to access non existent element");
  list < Element >::iterator pElem = elem_list.begin();
  for (int i = 0; i < num_elems; i++, pElem++)
    if (id == i)
      return *pElem;
}


const Element *ComplexType::getElement(const string & nam) 
{
  list < Element >::iterator pElem = elem_list.begin();
  while (pElem != elem_list.end())
    {
      if (pElem->getName() == nam)
        return &(*pElem);
      pElem++;
    }
  return 0;
}


Attribute & ComplexType::getAttribute(int id)
{
  if (id < 0 || id >= num_atts)
    error("tried to access non existent attributes");
  list < Attribute >::iterator pAttr = att_list.begin();
  for (int i = 0; i < num_atts; i++, pAttr++)
    if (id == i)
      return *pAttr;
}


const Attribute *ComplexType::getAttribute(const string & nam) 
{
    list < Attribute >::iterator pAttr = att_list.begin();
    while (pAttr != att_list.end())
    {
        if (pAttr->getName() == nam)
            return &(*pAttr);
        pAttr++;
    }
    return 0;
}


bool ComplexType::checkOccurrences(void)
{
    list < Element >::iterator pElem = elem_list.begin();
    while (pElem != elem_list.end())
    {
        if (pElem->nOccurrences < pElem->min())
            return false;
        pElem++;
    }
    return true;
}


void ComplexType::resetCounters(void)
{
  list < Element >::iterator pElem = elem_list.begin();
  while (pElem != elem_list.end())
    {
      pElem->nOccurrences = 0;
      pElem++;
    }
}


void ComplexType::error(string msg) const
{
  msg+=name;
  SchemaParserException spe(msg);
  throw spe;
  return;
}

void ComplexType::inheritBaseType()
{
  if(baseType.type  !=0)
    {
      ComplexType * cType   = (ComplexType*) baseType.type;

      //TODO copy only content model
      string tmp=name,ns=nsUri;
      int id=typeId;
  
      *this=*cType;

      name=tmp;
      nsUri=ns;
      typeId=id;
    }else {
      error("Base type unknown ");
      
    }
}
