/* 
 * wsdlpull - A C++ parser  for WSDL  (Web services description language)
 * Copyright (C) 2003 Vivek Krishna
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *
 */

#include <schemaparser/SchemaParser.h>
SchemaParser::SchemaParser(XmlPullParser * parser, string Uri,
                           ostream & log)
                           :xParser(parser),
                            m_tnsUri(Uri), 
                            logFile(log)
{
  m_elem_qualified = false;
  m_attr_qualified = false;
  resolveFwdRefs=true;
    
  nElems = 0;
  nAttrs = 0;
  nGroups = 0;
  nImportedSchemas = 0;
  m_deleteXmlParser=false;

}

SchemaParser::SchemaParser(string Uri, string tns , ostream & log )
             : logFile(log),xParser(0),m_tnsUri(tns)
{
  ifstream ifs;
  string fname;
 
  if(fetchUri(Uri,fname))
    {
      ifs.open(fname.c_str());
      xParser = new XmlPullParser(ifs);
      xParser->setFeature(FEATURE_PROCESS_NAMESPACES, true);
      xParser->require(XmlPullParser::START_DOCUMENT, "", "");
      while (!ifs.fail() && xParser->getEventType() != xParser->END_DOCUMENT)
        { 
          xParser->nextTag();
          if (xParser->getEventType() == xParser->START_TAG &&
              xParser->getName() == "schema")
            {
              m_deleteXmlParser=true;
              m_tnsUri=tns;
              break; 
            }
        }

    }
  if(!m_deleteXmlParser) //something wron while opening the schema file
    {
      delete xParser;
      xParser=0;
    }
      
}

SchemaParser::~SchemaParser()
{

  //clear the Types table
  m_typesTable.clean();
  int i = 0;
  for (i = 0; i < nElems; i++)
    delete m_glb_elem[i];
  for (i = 0; i < nAttrs; i++)
    delete m_glb_attr[i];
  if(m_deleteXmlParser)
    delete xParser;

}


/*
 * Parses an schema definition.
 * This is the main entry method for the schema parser
 */
bool SchemaParser::parseSchemaTag()
{
  int i = 0;
  if(!xParser)
    return false;
  while (xParser->getEventType() != xParser->START_TAG)
    xParser->next();
  xParser->require(xParser->START_TAG, SchemaUri, "schema");
  int attcnt = xParser->getAttributeCount();

  //parse the schema tag's attributes
  for (i = 0; i < attcnt; i++)
    {
      string attName = xParser->getAttributeName(i);
      if ("targetNamespace" == attName)
        //store the tns URI
        m_tnsUri = xParser->getAttributeValue(i);
      if ("elementFormDefault" == attName)
        {
          if (xParser->getAttributeValue(i) == "unqualified")
            m_elem_qualified = false;

          else if (xParser->getAttributeValue(i) == "qualified")
            m_elem_qualified = true;
        }
      if ("attributeFormDefault" == attName)
        {
          if (xParser->getAttributeValue(i) == "unqualified")
            m_attr_qualified = false;

          else if (xParser->getAttributeValue(i) == "qualified")
            m_attr_qualified = true;
        }
    }
  for (i = xParser->getNamespaceCount(xParser->getDepth()) - 1;
       i > xParser->getNamespaceCount(xParser->getDepth() - 1) - 1; i--)
    if (xParser->getNamespaceUri(i) == m_tnsUri)
      m_tnsPrefix = xParser->getNamespacePrefix(i);
  m_typesTable.setTargetNamespace(m_tnsUri);
  try
    {

      do
        {
          xParser->nextTag();
          if (xParser->getEventType() == xParser->END_TAG)
            {
              if (xParser->getName() == "schema")
                break;
              while (xParser->getEventType() != xParser->START_TAG)
                xParser->nextTag();
            }

          /*
            This is the main loop
            Depending on the tag encountered call the appropriate routines
            Schema elements visible at this level
            1. Global  Element declarations
            2. Global attribute declarations
            3. Complex type and Simple type declarations

          */
          string elemName = xParser->getName();
          if (elemName == "element")
            parseElement();

          else if (elemName == "complexType")
            {
              XSDType *t = parseComplexType();
              m_typesTable.addType(t);
            }

          else if (elemName == "simpleType")
            {
              XSDType *t = parseSimpleType();
              m_typesTable.addType(t);
            }

          else if (elemName == "attribute")
            parseAttribute();

          else if (elemName == "annotation")
            parseAnnotation();

          else if (elemName == "import")       
            parseImport();
          else if (elemName=="include")
            parseInclude();
          else
            break;
        }
      while (true);
      if (nImportedSchemas > 0 && m_typesTable.detectUndefinedTypes())
        {
          m_typesTable.printUndefinedTypes(logFile);
          error("Undefined Types in namespace "+m_tnsUri);
        }
      if(shouldResolve())
        {
            
          resolveForwardElementRefs();
          resolveForwardAttributeRefs();
        }
        
    }
  catch(SchemaParserException spe)
    {
      spe.line = xParser->getLineNumber();
      spe.col = xParser->getColumnNumber();

      cout << spe.description << " at " 
           << spe.line << ":" << spe.col 
           << endl;

      return false;
    }
  return true;
}


void SchemaParser::parseAnnotation()
{

  do
    {
      xParser->nextToken();
      if (xParser->getEventType() == xParser->END_TAG
          && xParser->getName() == "annotation")
        break;
    }
  while (true);
}


XSDType *SchemaParser::parseComplexType()
{
  ComplexType *newType = new ComplexType(m_tnsUri);
  int attcnt = xParser->getAttributeCount();
  for (int i = 0; i < attcnt; i++)
    {
      if ("name" == xParser->getAttributeName(i))
        newType->setName(xParser->getAttributeValue(i));
      if ("mixed" == xParser->getAttributeName(i))
        newType->setContentModel(newType->MIXED);
    }

  do
    {

      //begin parsing the complex type's children
      xParser->nextTag();
      if (xParser->getEventType() == xParser->END_TAG)
        {
          if (xParser->getName() == "complexType")
            break;

          //if an end tag is seen proceed till next start tag
          while (xParser->getEventType() != xParser->START_TAG)
            xParser->nextTag();
        }
      string elemName = xParser->getName();
      if (elemName == "all")
        all(newType);

      else if (elemName == "sequence")
        cs(newType);

      else if (elemName == "choice")
        cs(newType);

      else if (elemName == "attribute")
        addAttribute(newType);

      else if (elemName == "anyAttribute")
        addAnyAttribute(newType);

      else if (elemName == "complexContent")
        parseComplexContent(newType);

      else if (elemName == "simpleContent")
        parseSimpleContent(newType);

      else if (xParser->getName() == "annotation")
        parseAnnotation();

      else
        error("<complexType>: Syntax error");
    }
  while (true);
  return newType;
}


int SchemaParser::parseGroup()
{
  int min = 1, max = 1;
  string tmp, name, ref;
  ComplexType *ct = new ComplexType(m_tnsUri);
  if (xParser->getName() != "group")
    error("<group>Invalid State");
  tmp = xParser->getAttributeValue("", "minOccurs");
  if (!tmp.empty())
    min = parseInt(tmp);
  tmp = xParser->getAttributeValue("", "maxOccurs");
  if (!tmp.empty())
    {
      if ("unbounded" == tmp)
        max = UNBOUNDED;

      else
        max = parseInt(tmp);
    }
  name = xParser->getAttributeValue("", "name");
  ref = xParser->getAttributeValue("", "ref");
  if (!ref.empty())
    {
      Qname gRef(ref);
      return 0;

      //      return getGroupId(gRef);TODO
    }
  xParser->nextTag();
  while (xParser->getName() == "annotation")
    {
      parseAnnotation();
      xParser->nextTag();
    }
  string elemName = xParser->getName();
  if (elemName == "all")
    all(ct);

  else if (elemName == "sequence")
    cs(ct);

  else if (elemName == "choice")
    cs(ct);
  m_groups[nGroups].name = name;
  m_groups[nGroups].min = min;
  m_groups[nGroups].max = max;
  m_groups[nGroups].mgs = ct;
  return nGroups++;
}


void SchemaParser::all(ComplexType * ct)
{
  int min = 1, max = 1;
  string tmp;
  if (xParser->getName() != "all")
    error("<all> :Invalid State");
  tmp = xParser->getAttributeValue("", "minOccurs");
  if (!tmp.empty())
    min = parseInt(tmp);
  tmp = xParser->getAttributeValue("", "maxOccurs");
  if (!tmp.empty())
    max = parseInt(tmp);
  ct->setCompositor(ct->ALL, true, min, max);
  xParser->nextTag();
  while (xParser->getName() == "annotation")
    {
      parseAnnotation();
      xParser->nextTag();
    }
  while (xParser->getName() == "element")
    {
      addElement(ct);
      xParser->nextTag();
    }
  if (xParser->getName() == "all"
      && xParser->getEventType() == xParser->END_TAG)
    ct->setCompositor(ct->ALL, false);

  else
    error("<all>:Syntax error ");
  return;
}


void SchemaParser::cs(ComplexType * ct)
{
  int min = 1, max = 1;
  string tmp;
  while (xParser->getName() == "annotation")
    {
      parseAnnotation();
      xParser->nextTag();
    }
  if (xParser->getName() != "choice" && xParser->getName() != "sequence")
    error("<cs>:Invalid State");
  tmp = xParser->getAttributeValue("", "minOccurs");
  if (!tmp.empty())
    min = parseInt(tmp);
  tmp = xParser->getAttributeValue("", "maxOccurs");
  if (!tmp.empty())
    {
      if ("unbounded" == tmp)
        max = UNBOUNDED;

      else
        max = parseInt(tmp);
    }
  if (xParser->getName() == "choice")
    ct->setCompositor(ct->CHOICE, true, min, max);

  else
    ct->setCompositor(ct->SEQ, true, min, max);
  xParser->nextTag();
  while (!((xParser->getEventType() == xParser->END_TAG) &&
           (xParser->getName() == "choice"
            || xParser->getName() == "sequence")))
    {
      if (xParser->getName() == "element")
        addElement(ct);

      else if (xParser->getName() == "any")
        addAny(ct);

      else if (xParser->getName() == "choice")
        cs(ct);

      else if (xParser->getName() == "sequence")
        cs(ct);

      else
        error("<cs>:Syntax Errror");
      xParser->nextTag();
    }
  if (xParser->getName() == "choice")
    ct->setCompositor(ct->CHOICE, false);

  else
    ct->setCompositor(ct->SEQ, false);

  //close the compositor
  return;
}


//Adds an element to a complex type
void SchemaParser::addElement(ComplexType * cType)
{
  string name, fixedVal, defaultVal;
  int type_id = 0, min = 1, max = 1, attcnt;
  Qname refName;
  bool qualified = false, use = true, added = false, nill = false;
  XSDType *elemType;
  attcnt = xParser->getAttributeCount();
  for (int i = 0; i < attcnt; i++)
    {
      string attName = xParser->getAttributeName(i);
      if ("name" == attName)
        name = xParser->getAttributeValue(i);

      else if ("type" == attName)
        {
          Qname typeName(xParser->getAttributeValue(i));
          if (type_id > 0)
            error
              ("<element> : type and ref are mutually exclusive in element decl");
          typeName.setNamespace(xParser->
                                getNamespace(typeName.getPrefix()));
          type_id = getTypeId(typeName, true);
          if (type_id == 0)
            error("<element>:Could not resolve type " +
                  typeName.getNamespace() + ":" +
                  typeName.getLocalName());
        }

      else if ("form" == attName)
        {
          if ("qualified" == xParser->getAttributeValue(i))
            qualified = true;

          else if ("unqualified" == xParser->getAttributeValue(i))
            qualified = false;

          else
            error("<element>:Invalid value for form in element " +
                  name);
        }

      else if ("ref" == attName)
        {
          if (!name.empty())
            error
              ("<element>:name and ref are mutually exclusive in element decl");
          if (type_id > 0)
            error
              ("<element>:type and ref are mutually exclusive in element decl");
          refName = xParser->getAttributeValue(i);
          refName.setNamespace(xParser->getNamespace(refName.getPrefix()));
          Element *e=0;
          if(refName.getNamespace()==m_tnsUri){
             e = getElement(getElementId(refName));
          }else{
            int i=checkImport(refName.getNamespace());
            if(m_ImportedSchemas[i].sParser)
              e=m_ImportedSchemas[i].sParser->getElement(refName);
          }
          
          
          if (e == 0)
            {
              added = true;
              m_ForwardElementRef.push_back(refName);
            }

          else
            {
              name = e->getName();
              type_id = e->getType();
              qualified = e->isQualified();
              defaultVal = e->defaultVal();
              fixedVal = e->fixedVal();
            }
        }
      
      else if ("minOccurs" == attName)
        min = parseInt(xParser->getAttributeValue(i), 10);

      else if ("maxOccurs" == attName)
        {
          if ("unbounded" == xParser->getAttributeValue(i))
            max = UNBOUNDED;

          else
            max = parseInt(xParser->getAttributeValue(i), 10);
          if (max == -1)                        //invalid value for maxOccurs
            error("<element>:Invalid value for maxOccurs");
        }

      else if ("default" == attName)
        {
          if (fixedVal.empty())
            defaultVal = xParser->getAttributeValue(i);

          else
            error("<element>:fixed and default cannot occur together");
        }

      else if ("fixed" == attName)
        {
          if (defaultVal.empty())
            fixedVal = xParser->getAttributeValue(i);

          else
            error("<element>:fixed and default cannot occur together");
        }

      else if ("substitutionGroup" == attName)
        {

          //do nothing
        }

      else if ("nillable" == attName)
        {

          //a nillable element need not have a type ,so set it to anyType id if nop type was given
          nill = true;
        }

      else
        error("<element>:Unsupported Attribute");
    }

  do
    {
      xParser->nextTag();
      if (xParser->getEventType() == xParser->END_TAG)
        {
          if (xParser->getName() == "element")
            break;

          //if an end tag is seen proceed till next start tag
          while (xParser->getEventType() != xParser->START_TAG)
            xParser->nextTag();
        }
      if (xParser->getName() == "complexType")
        {
          elemType = parseComplexType();

          //create an anonymous type
          ComplexType *ct = (ComplexType *) elemType;
          if (ct->getNumChildren() == 1
              && ct->getChildType(0) == XSDType::ANY
              && ct->getChildName(0) == "any")
            {

              //if the complex type is <any> then we dont need a type for it.
              //make the parent's type as XSDType::ANY
              delete ct;
              type_id = XSDType::ANY;
            }

          else
            type_id = m_typesTable.addType(elemType);
        }

      else if (xParser->getName() == "simpleType")
        {
          elemType = parseSimpleType();

          //create an anonymous type
          type_id = m_typesTable.addType(elemType);
        }

      else if (xParser->getName() == "annotation")
        parseAnnotation();

      else
        error("<element>:Syntax error or unkown tag");
    }
  while (true);
  if (nill && type_id == 0)
    type_id = XSDType::ANYTYPE;
  if (!added)
    cType->addElement(name, type_id, min, max, qualified, defaultVal,
                      fixedVal);

  else
    cType->addElementRef(refName, min, max);
}


void SchemaParser::addAny(ComplexType * cType)
{
  string ns, any("any");
  int type_id = XSDType::ANY, min = 1, max = 1, attcnt;
  bool qualified = false, use = true;
  XSDType *elemType;

  //note  processContents=lax .
  attcnt = xParser->getAttributeCount();
  for (int i = 0; i < attcnt; i++)
    {
      string attr = xParser->getAttributeName(i);
      if ("namespace" == attr)
        ns = xParser->getAttributeValue(i);

      else if ("minOccurs" == attr)
        min = parseInt(xParser->getAttributeValue(i), 10);

      else if ("maxOccurs" == attr)
        {
          if ("unbounded" == xParser->getAttributeValue(i))
            max = UNBOUNDED;

          else
            max = parseInt(xParser->getAttributeValue(i), 10);
          if (max == -1)                        //invalid value for maxOccurs
            error("<element>:Invalid value for maxOccurs");
        }

      else if ("processContents" == attr || "id" == attr)
        {

          //do nothing
        }

      else
        error("<any>:Unsupported Attribute");
    }
  cType->addElement(any, type_id, min, max, false, ns);
  xParser->nextTag();
}


void SchemaParser::addAnyAttribute(ComplexType * cType)
{
  string ns, anyAttribute("anyAttribute");
  int type_id = XSDType::ANY, min = 1, max = 1, attcnt;
  bool qualified = false, use = true;
  XSDType *elemType;

  //note  processContents=lax .
  attcnt = xParser->getAttributeCount();
  for (int i = 0; i < attcnt; i++)
    {
      string attr = xParser->getAttributeName(i);
      if ("namespace" == attr)
        ns = xParser->getAttributeValue(i);

      else if ("processContents" == attr || "id" == attr)
        {

          //do nothing
        }

      else
        error("<anyAttribute>:Unsupported Attribute");
    }
  cType->addAttribute(anyAttribute, type_id, false, ns);
  xParser->nextTag();
  while (xParser->getName() == "annotation")
    {
      parseAnnotation();
      xParser->nextTag();
    }
}


//This function parses and attribute and adds it to the complexType
void SchemaParser::addAttribute(ComplexType * cType)
{
  string name, fixedVal, defaultVal;
  int type_id = 0, attcnt;
  bool qualified = false, use = false, added = false;
  Qname refAttribute;
  attcnt = xParser->getAttributeCount();
  for (int i = 0; i < attcnt; i++)
    {
      string attName = xParser->getAttributeName(i);
      if ("name" == attName)
        name = xParser->getAttributeValue(i);

      else if ("type" == attName)
        {
          if (type_id > 0)
            error
              ("<attribute>:type and ref are mutually exclusive in element decl");
          Qname typeName(xParser->getAttributeValue(i));
          typeName.setNamespace(xParser->
                                getNamespace(typeName.getPrefix()));
          type_id = getTypeId(typeName, true);
          if (type_id == 0)
            error("<attribute>:Could not resolve type " +
                  typeName.getNamespace() + ":" +
                  typeName.getLocalName());
        }

      else if ("form" == attName)
        {
          if ("qualified" == xParser->getAttributeValue(i))
            qualified = true;

          else if ("unqualified" == xParser->getAttributeValue(i))
            qualified = false;

          else
            error("<attribute>:Invalid value for form");
        }

      else if ("ref" == attName)
        {
          if (!name.empty())
            error
              ("<attribute>:name and ref are mutually exclusive in element decl");
          if (type_id > 0)
            error
              ("<attribute>:type and ref are mutually exclusive in element decl");
          refAttribute = xParser->getAttributeValue(i);
          refAttribute.setNamespace(xParser->getNamespace(refAttribute.getPrefix()));
          Attribute *a =0;
          if(refAttribute.getNamespace()==m_tnsUri){
            
            a=getAttribute(getAttributeId(refAttribute));
          }else{
            int i=checkImport(refAttribute.getNamespace());
            if(m_ImportedSchemas[i].sParser){
              a=m_ImportedSchemas[i].sParser->getAttribute(refAttribute);
            }
            else
              a=0;
            
          }
          
          if (a == 0)
            {
              added = true;
              m_ForwardAttributeRef.push_back(refAttribute);
            }

          else
            {
              name = a->getName();
              type_id = a->getType();
              qualified = a->isQualified();
              defaultVal = a->defaultVal();
              fixedVal = a->fixedVal();
            }
        }

      else if ("default" == attName)
        {
          if (fixedVal.empty())
            defaultVal = xParser->getAttributeValue(i);

          else
            error
              ("<attribute>:fixed and default cannot occur together");
        }

      else if ("fixed" == attName)
        {
          if (defaultVal.empty())
            fixedVal = xParser->getAttributeValue(i);

          else
            error
              ("<attribute>:fixed and default cannot occur together");
        }

      else if ("use" == attName)
        {
          if (xParser->getAttributeValue(i) == "optional")
            use = false;

          else if (xParser->getAttributeValue(i) == "required")
            use = true;

          else
            error("<attribute>:unknown attribute value");
        }

      else
        error("<attribute>:Unsupported attributes");
    }

  //Now parse the children of the attribute tag viz simpleType
  do
    {
      xParser->nextTag();
      if (xParser->getEventType() == xParser->END_TAG)
        {
          if (xParser->getName() == "attribute")
            break;

          //if an end tag is seen proceed till next start tag
          while (xParser->getEventType() != xParser->START_TAG)
            xParser->nextTag();
        }

      else if (xParser->getName() == "simpleType")
        {
          XSDType *elemType = parseSimpleType();

          //create an anonymous type
          type_id = m_typesTable.addType(elemType);
        }

      else if (xParser->getName() == "annotation")
        parseAnnotation();

      else
        error("<attribute>:Syntax error or unkown tag");
    }
  while (true);
  if (!added)
    cType->addAttribute(name, type_id, qualified, defaultVal, fixedVal,
                        use);

  else
    cType->addAttributeRef(refAttribute, qualified, use);
}


XSDType *SchemaParser::parseSimpleType()
{
  SimpleType *st = new SimpleType(m_tnsUri);
  int basetype_id = 0;
  int attcnt;
  attcnt = xParser->getAttributeCount();
  for (int i = 0; i < attcnt; i++)
    {
      if ("name" == xParser->getAttributeName(i))
        st->setName(xParser->getAttributeValue(i));

      else
        error("<simpleType> :" + xParser->getAttributeName(i) +
              ":Unknown/Unsupported  attribute ");
    }

  do
    {
      xParser->nextTag();
      if (xParser->getEventType() == xParser->END_TAG)
        {
          if (xParser->getName() == "simpleType")
            break;

          //if an end tag is seen proceed till next start tag
          while (xParser->getEventType() != xParser->START_TAG)
            xParser->nextTag();
        }
      if (xParser->getName() == "restriction")
        {
          attcnt = xParser->getAttributeCount();
          for (int i = 0; i < attcnt; i++)
            {
              if ("base" == xParser->getAttributeName(i))
                {
                  Qname typeName(xParser->getAttributeValue(i));
                  typeName.setNamespace(xParser->
                                        getNamespace(typeName.
                                                     getPrefix()));
                  st->setBaseType(basetype_id =
                                  getTypeId(typeName, true));
                  if (basetype_id == 0)
                    error("<simpleType>:" +
                          xParser->getAttributeValue(i) +
                          ":Unkown base type ");
                }

              else
                error("<simpleType>:" + xParser->getAttributeName(i) +
                      ":Unknown/Unsupported  attribute for <restriction>");
            }
          parseRestriction(st);
        }

      else if (xParser->getName() == "union"
               || xParser->getName() == "list")
        error("<simpleType>::Union and List  not supported");

      else if (xParser->getName() == "annotation")
        parseAnnotation();

      else
        error("<simpleType>:Syntax error");
    }
  while (true);
  return st;
}


void SchemaParser::parseRestriction(SimpleType * st)
{
  if (st->getBaseType() == 0)
    error("<restriction>:unkown BaseType");

  do
    {
      xParser->nextTag();
      if (xParser->getEventType() == xParser->END_TAG)
        {
          if (xParser->getName() == "restriction")
            break;
          else
            xParser->nextTag();
          if (xParser->getName() == "restriction"
              && xParser->getEventType() == xParser->END_TAG)
            break;
        }
      while (xParser->getName() == "annotation")
        {
          parseAnnotation();
          xParser->nextTag();
        }
      //TODO check the parser states here,no simpleType allowed inside the restriction
      if (!st->isvalidFacet(xParser->getName()))
        //This function also sets the facet if valid
        error("<restriction>:" + xParser->getName() +
              " is not a valid facet for the simple type");
      st->setFacetValue(xParser->getAttributeValue("", "value"));
    } while (true);
}


void SchemaParser::parseComplexContent(ComplexType * ct)
{
  int attcnt = xParser->getAttributeCount();
  int basetype_id = 0, i = 0;
  Qname typeName;
  for (i = 0; i < attcnt; i++)
    {
      if ("mixed" == xParser->getAttributeName(i))
        if (xParser->getAttributeValue(i) == "true")
          error("<complexContent>: No support for mixed=true");
    }
  ct->setContentModel(ct->COMPLEX);
  xParser->nextTag();
  if (xParser->getName() == "restriction")
    {
      attcnt = xParser->getAttributeCount();
      for (i = 0; i < attcnt; i++)
        {
          if ("base" == xParser->getAttributeName(i))
            {
              typeName = xParser->getAttributeValue(i);
              typeName.setNamespace(xParser->
                                    getNamespace(typeName.getPrefix()));
            }
        }
      ct->setBaseType(getTypeId(typeName, true),
                      ComplexType::Restriction,
                      getType(typeName));
      

    }
  else if (xParser->getName() == "extension")
    {
      attcnt = xParser->getAttributeCount();
      for (i = 0; i < attcnt; i++)
        {
          if ("base" == xParser->getAttributeName(i))
            {
              typeName = xParser->getAttributeValue(i);
              typeName.setNamespace(xParser->
                                    getNamespace(typeName.getPrefix()));
            }
        }
      ct->setBaseType(getTypeId(typeName, true),
                      ComplexType::Extension,
                      getType(typeName));
      
    }
  
  xParser->nextTag();

  //if the base soapenc:Array ,then read only the arrayType attribute and nothing else
  if (typeName.getNamespace() == soapEncUri
      && typeName.getLocalName() == "Array")
    {
      ct->array(true);
      Qname arrayType(xParser->
                      getAttributeValue(wsdlUri, "arrayType"));
      arrayType.setNamespace(xParser->
                             getNamespace(arrayType.getPrefix()));
      ct->addElement(arrayType.getLocalName(),
                     getTypeId(arrayType, true), 0, UNBOUNDED);
    }else{
      
      if (xParser->getName() == "all")
        all(ct);
      
      else if (xParser->getName() == "sequence")
        cs(ct);
      
      else if (xParser->getName() == "choice")
        cs(ct);
      
      else if (xParser->getName() == "attribute")
        addAttribute(ct);
  
      else if (xParser->getName() == "anyAttribute")
        addAnyAttribute(ct);
    }
  
  do
    {
      xParser->nextTag();
      if (xParser->getEventType() == xParser->END_TAG)
        if ((xParser->getName() == "restriction" ||
             xParser->getName() == "extension") ) 
          break;
    }
  while (true);

  xParser->nextTag();
}


void SchemaParser::parseSimpleContent(ComplexType * ct)
{
  int attcnt = xParser->getAttributeCount();
  if (attcnt > 0)
    error
      ("<simpleContent> :  No attributes are currently supported for <simpleContent>");
  ct->setContentModel(ct->SIMPLE);
  xParser->nextTag();
  if (xParser->getName() == "restriction")
    {
      SimpleType *st = new SimpleType(m_tnsUri);
      attcnt = xParser->getAttributeCount();
      int basetype_id = 0;
      for (int i = 0; i < attcnt; i++)
        {
          if ("base" == xParser->getAttributeName(i))
            {
              Qname typeName(xParser->getAttributeValue(i));
              typeName.setNamespace(xParser->
                                    getNamespace(typeName.getPrefix()));
              st->setBaseType(basetype_id = getTypeId(typeName, true));
              if (basetype_id == 0)
                error("<simpleContent> :" +
                      xParser->getAttributeValue(i) +
                      ":Unkown base type ");
            }

          else
            error("<simpleContent> :" + xParser->getAttributeName(i) +
                  ":Unknown/Unsupported  attribute ");
        }
      xParser->nextTag();
      parseRestriction(st);
      int typeId = m_typesTable.addType(st);
      if (typeId == 0)
        error("Could not add type in types table");
      ct->setContentTypeId(typeId);
    }

  else if (xParser->getName() == "extension")
    {

      //This extension does not use the full model that can come in
      //ComplexContent .It uses the simple model.no particle allowed ,only attributes
      attcnt = xParser->getAttributeCount();
      int basetype_id = 0;
      for (int i = 0; i < attcnt; i++)
        {
          if ("base" == xParser->getAttributeName(i))
            {
              Qname typeName(xParser->getAttributeValue(i));
              typeName.setNamespace(xParser->
                                    getNamespace(typeName.getPrefix()));
              ct->setContentTypeId(basetype_id =
                                   getTypeId(typeName, true));
              if (basetype_id == 0)
                error("<simpleContent> :" +
                      xParser->getAttributeValue(i) +
                      ":Unkown base type ");
            }

          else
            error("<simpleContent> :" + xParser->getAttributeName(i) +
                  ":Unknown/Unsupported  attribute ");
        }
      xParser->nextTag();
      while (xParser->getName() == "attribute")
        {
          addAttribute(ct);
          xParser->nextTag();
        }
      if (!
          (xParser->getName() == "extension"
           && xParser->getEventType() == xParser->END_TAG))
        error("<simpleContent> :Syntax error :extension");
    }
  xParser->nextTag();
  if (!
      (xParser->getName() == "simpleContent"
       && xParser->getEventType() == xParser->END_TAG))
    error("<simpleContent> :Syntax error ");
}


bool SchemaParser::parseInclude()
{
  ifstream xsdStream;
  string fname;
  string attVal = xParser->getAttributeValue("", "schemaLocation");

  if(!attVal.empty())
    {
      if(fetchUri(attVal,fname))
        {
          /*
           * If the schema definition was retrieved successfully 
           * process it and add all type definitions and
           * declaration to the current namespace
           */
          xsdStream.open(fname.c_str());
            
          XmlPullParser * xpp = new XmlPullParser(xsdStream);
          XmlPullParser * tmpXparser=xParser;
          xParser=xpp;

          xParser->setFeature(FEATURE_PROCESS_NAMESPACES, true);
          xParser->require(XmlPullParser::START_DOCUMENT, "", "");
          while (xParser->getEventType() != xParser->END_DOCUMENT)
            { 
              xParser->nextTag();
              if (xParser->getEventType() == xParser->START_TAG &&
                  xParser->getName() == "schema")
                {
                  resolveFwdRefs=false;
                   
                  if(!parseSchemaTag())
                    error("Error while parsing the included schema " + attVal);
                  else{

                    resolveFwdRefs=true;  
                    break;
                  }
                  

                   
                }
            }
          xParser=tmpXparser;
          delete xpp;
        }else
          {
            error("Error while opening the included schema " + attVal);
          }
    }else
      {
        error("schemaLocation is a required attribute for <include>");
      }
          
  xParser->nextTag();
  return true;
}

bool SchemaParser::parseImport()
{
  int attcnt = xParser->getAttributeCount();
  Qname typeName;
  string xsdFile;
  string ns = xParser->getAttributeValue("", "namespace");
  string loc=xParser->getAttributeValue("", "schemaLocation");
   
  if(!loc.empty())
    {
      if(fetchUri(loc,xsdFile))
        {
          /*
           * If the schema definition was retrieved successfully 
           * process it and add it to list of imported schemas
           */
          SchemaParser *sp = new SchemaParser(xsdFile,ns);
          if(sp->parseSchemaTag())
            addImport(sp);
          else
            error("Error while parsing imported namespace "+ns);
               
        }
    }
  else // if no location is mentioned ,just add the namespace,types will be resolved later
    addImport(ns);
    
  xParser->nextTag();
  return true;
}

bool SchemaParser::isBasicType(int sType) const
{
  if (sType > XSDType::ANYURI)
    return false;

  else
    return true;
}


//Parse global element decrations
void SchemaParser::parseElement()
{
    ComplexType *tmpCt = new ComplexType(m_tnsUri);
    addElement(tmpCt);
    Element *e = new Element();
    *e = tmpCt->getElement(0);
    delete tmpCt;
    if (nElems >= 40)
        error("Max limit for number of global elements");
    m_glb_elem[nElems++] = e;
}


// Parse global attrribute decrations
void SchemaParser::parseAttribute()
{
  ComplexType *tmpCt = new ComplexType(m_tnsUri);
  addAttribute(tmpCt);
  Attribute *a = new Attribute();
  *a = tmpCt->getAttribute(0);
  delete tmpCt;
  if (nAttrs >= 16)
    error("Max limit for number of global attributes");
  m_glb_attr[nAttrs++] = a;
}

//store a reference to an imported element reference
int  SchemaParser::addExternalElement(const string & name ,int localTypeId)
{
  Element *e = new Element(name,localTypeId);
  if (nElems >= 40)
    error("Max limit for number of global elements");
  m_glb_elem[nElems++] = e;
  return nElems-1;
}


//This function gets the id of a type whose Qname is passed.
//The Qname struct if it has a prefix must either be a valid namespace
//default is http://www.w3.org/2001/XMLSchema
//This function has two modes
//if 'create' is true a new type is created (for fwd references)
//otherwise the existing list of parsed types is used for resolution

int SchemaParser::getTypeId( Qname &  type, bool create)
{
  string typens = type.getNamespace();
  if (typens.empty())
    type.setNamespace(typens = m_tnsUri);
  if (typens == m_tnsUri || typens == SchemaUri)
    return m_typesTable.getTypeId(type, create);

  else
    {
      //postpone resolution till matchExtRefs is called
      if (nImportedSchemas == 0)                
        return m_typesTable.addExternalTypeId(type, 0);

      //search in the array of imported schemas
      int typeId = 0;
      for (int i = 0; i < nImportedSchemas; i++)
        {
          if ( m_ImportedSchemas[i].ns == type.getNamespace())
            {
              if(m_ImportedSchemas[i].sParser )
		{
                  typeId = m_ImportedSchemas[i].sParser->getTypeId(type, false);
                  if (typeId)
                    return m_typesTable.addExternalTypeId(type, (XSDType *) m_ImportedSchemas[i].sParser->getType(typeId));
                  else
                    return 0;
		}
            }
        }
      return m_typesTable.addExternalTypeId(type, 0);
    }
}


//resolves any external references with  the imported schemas
//This method must be called to ensure resolution of all types
bool SchemaParser::finalize(void)
{
  int unresolved=m_typesTable.getNumExtRefs();
  if(unresolved==0)
    return true;

  for (int i = 0; i < unresolved; i++)
    {
      Qname & type = m_typesTable.getExtRefName(i);
      int localId = m_typesTable.getExtRefType(i);

      //search in the array of imported schemas
      int typeId = 0;
      for (int n = 0; n < nImportedSchemas; n++)
        {
          if (m_ImportedSchemas[n].ns == type.getNamespace())
            {
	      if(m_ImportedSchemas[n].sParser){
                typeId = m_ImportedSchemas[n].sParser->getTypeId(type);
                if (typeId != 0)
                  m_typesTable.addExtType((XSDType *) m_ImportedSchemas[n].sParser->getType(typeId),
                                          localId);
	      }
            }
        }
    }
  if (m_typesTable.detectUndefinedTypes())
    {
      m_typesTable.printUndefinedTypes(logFile);
      return false;
    }

  else{
    
    return true;
  }
  
}


//resolves any forward references of the kind<element ref=Qname... >
void SchemaParser::resolveForwardElementRefs()
{
  if (m_ForwardElementRef.empty())
    return;
  for (list < Qname >::iterator pQnames = m_ForwardElementRef.begin();
       pQnames != m_ForwardElementRef.end(); pQnames++)
    {
      Element *e = getElement(getElementId(*pQnames));
      if (e)
        m_typesTable.resolveForwardElementRefs(pQnames->getLocalName(),*e);
      else error("Could not resolve element reference "+pQnames->getLocalName());
           
    }
}


void SchemaParser::resolveForwardAttributeRefs()
{
  if (m_ForwardAttributeRef.empty())
    return;
  for (list < Qname >::iterator pQnames = m_ForwardAttributeRef.begin();
       pQnames != m_ForwardAttributeRef.end(); pQnames++)
    {
      Attribute *a = getAttribute(getAttributeId(*pQnames));
      if (a)
        m_typesTable.resolveForwardAttributeRefs(pQnames-> getLocalName(), *a);
      else error("Could not resolve attribute reference  "+pQnames->getLocalName());
    }
}


//get the element id of a globally declared element
int SchemaParser::getElementId(Qname & type)
{
  string typens = type.getNamespace();
  if (typens.empty())
    type.setNamespace(typens = m_tnsUri);
  if (typens== m_tnsUri || typens == SchemaUri)
    {
      int i = 0;

      //check if it is a global element
      for (i = 0; i < nElems; i++)
        if (m_glb_elem[i]->getName() == type.getLocalName())
          return i;
      return -1;
    }
  else
    { //search imported namespaces
      for (int i = 0; i < nImportedSchemas; i++)
        {
          if ( m_ImportedSchemas[i].ns == typens)
            {
              if(m_ImportedSchemas[i].sParser )
		{
                  int typeId = m_ImportedSchemas[i].sParser->getElementType(type);
                
                  if (typeId){
                    //found the definition
                    int localTypeId= m_typesTable.addExternalTypeId(type, 
                                                                    (XSDType *) m_ImportedSchemas[i].sParser->
                                                                    getType(typeId));
                    return addExternalElement(type.getLocalName(),
                                              localTypeId);
                  }
                  else
                    return -1;
		}
            }
        }
    }
  return -1;
       
}


//returns the type of the element
int SchemaParser::getElementType(Qname & type)
{
  int elemId = getElementId(type);
  if (elemId == -1)
    return 0;
  Element *e = getElement(elemId);
  if (e != 0)
    return e->getType();

  else
    return 0;
}

Element *SchemaParser::getElement(Qname & element)const
{

  string typens = element.getNamespace();
  if (typens.empty())
    element.setNamespace(typens = m_tnsUri);
  if (typens== m_tnsUri || typens == SchemaUri)
    {
      int i = 0;

      //check if it is a global element
      for (i = 0; i < nElems; i++)
        if (m_glb_elem[i]->getName() == element.getLocalName())
          return m_glb_elem[i];
      return 0;
    }
  return 0;
  
}


Element *SchemaParser::getElement(int id) const
{
    if (id >= 0 && id < nElems)
        return m_glb_elem[id];

    else
        return 0;
}


//get the attribute id of a globally declared attribute
int SchemaParser::getAttributeId(Qname & type)const
{
  string typens = type.getNamespace();
  if (typens.empty())
    type.setNamespace(typens = m_tnsUri);
  if (typens != m_tnsUri && typens != SchemaUri)
    return -1;
  int i = 0;

  //check if it is a global attribute
  for (i = 0; i < nAttrs; i++)
    if (m_glb_attr[i]->getName() == type.getLocalName())
      return i;
  return -1;
}


//returns the type of the attribute
int SchemaParser::getAttributeType(Qname & type)
{
  int attId = getAttributeId(type);
  if (attId == -1)
    return 0;
  Attribute *a = getAttribute(attId);
  if (a != 0)
    return a->getType();

  else
    return 0;
}


Attribute *SchemaParser::getAttribute(int id) const
{
  if (id >= 0 && id < nAttrs)
    return m_glb_attr[id];

  else
    return 0;
}

Attribute *SchemaParser::getAttribute( Qname & attribute) const
{

  int id=getAttributeId(attribute);
  if(id!=-1){
    
    return m_glb_attr[id];
  }else{
  
    return 0;
  }
  
  
}


string SchemaParser::getTargetNamespace(void) const
{
    return m_tnsUri;
}


// Sun Workshop compiler gives an error 'inline', since it's already been
// called.  'inline' for a .cpp is probably not that useful anyway.  -KEC
const XSDType *SchemaParser::getType(int id) const
{
  return (const XSDType *) m_typesTable.getTypePtr(id);
}


const XSDType *SchemaParser::getType(const Qname & type ) 
{
  int id;
  Qname t =type;
  
  if((id=getTypeId(t,false))==0)
    return 0;
  else
    return (const XSDType *) m_typesTable.getTypePtr(id);
}


list < const XSDType *>*SchemaParser::getAllTypes() const
{
    list < const XSDType *>*pLTypes = new list < const XSDType * >;
    for (int i = 0; i < getNumTypes(); i++)
    {
        const XSDType *pType = getType(i + XSDType::ANYURI + 1);
        pLTypes->push_back(pType);
    }
    return pLTypes;
}


int SchemaParser::getNumTypes() const
{
    return m_typesTable.getNumTypes();
}


int SchemaParser::getNumElements() const
{
    return nElems;
}


int SchemaParser::getNumAttributes() const
{
    return nAttrs;
}

//To be deprecated
bool SchemaParser::addImports(SchemaParser ** schemaParsers, int n)
{
  if(n>max_imports)
    return false;
  for (int i=0;i<n ;i++)
    {
     m_ImportedSchemas[i].sParser = schemaParsers[i];
     m_ImportedSchemas[i].ns=schemaParsers[i]->getTargetNamespace();
    }
    

  nImportedSchemas = n;
}

bool SchemaParser::addImport(SchemaParser *sp)
{
  if(nImportedSchemas>=max_imports || sp==0 )
    return false;
   //check if the namespace is added in the import list
   int i= checkImport(sp->getTargetNamespace());

   m_ImportedSchemas[i].sParser=sp;
   m_ImportedSchemas[i].ns=sp->getTargetNamespace();
   //if this was a new import increment
   if(i==nImportedSchemas)
       nImportedSchemas++;
  return true;
}

int SchemaParser::checkImport(string nsp)
{
  for(int i=0;i<nImportedSchemas;i++)
  {
   if(m_ImportedSchemas[i].ns==nsp)
      return i;
  }
return nImportedSchemas;
}

bool SchemaParser::addImport(string ns,string location)
{
  if(nImportedSchemas>=max_imports)
    return false;
   int i= checkImport(ns);
   m_ImportedSchemas[i].sParser=0;
   m_ImportedSchemas[i].ns=ns;
   if(i==nImportedSchemas)
       nImportedSchemas++;
   if(location.empty())
     return true;
   string xsdFile;
   if(fetchUri(location,xsdFile))
   {
	    /*
             * If the schema definition was retrieved successfully 
             * process it and add it to list of imported schemas
             */
	    SchemaParser *sp = new SchemaParser(xsdFile,ns);
            if(sp->parseSchemaTag())
	      {
                m_ImportedSchemas[i].sParser=sp;
                return true;
	      }
              else return false;
   }
   else return false;

}


void SchemaParser::error(string mesg, int level)
{
    if (level == 0)
    {
        SchemaParserException spe(mesg + "\nError in SchemaParser\n");
        spe.line = xParser->getLineNumber();
        spe.col = xParser->getColumnNumber();
        throw spe;
    }

    else if (level == 1)
    {
        logFile << "Warning @" << xParser->
            getLineNumber() << ":" << xParser->
            getColumnNumber() << dbsp << mesg << endl;
    }
}
