/* 
 * wsdlpull - A C++ parser  for WSDL  (Web services description language)
 * Copyright (C) 2003 Vivek Krishna
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *
 */
#include <sstream>
#include <schemaparser/SchemaValidator.h>

/*
  This file contains methods of the schema parser which
  validate an incoming Xml stream against a given type
  whose schema has been processed by this SchemaParser
  instance

*/
SchemaValidator::SchemaValidator(const SchemaParser * sp)
{
  sParser = sp;
}


SchemaValidator::~SchemaValidator()
{
}


/*
Main entry method for validation
Inputs
1. XmlStream ,xpp.getName() muct return the name of the  type
    which must be validated
2. typeId of the type against which this stream must be
   validated against
3.An Input type container  (can be 0)
*/
TypeContainer *SchemaValidator::validate(XmlPullParser * xpp, int typeId,
                                         TypeContainer * ipTc)
{
  TypeContainer *t;
  string elemName = xpp->getName();
  if (!ipTc)
    t = new TypeContainer(typeId, sParser);

  else
    t = ipTc;
  if (t->getTypeId() != typeId)
    error("Fatal error");
  if (sParser->getType(typeId) == 0
      || sParser->getType(typeId)->isSimple())
    {

      string val;
      xpp->nextToken();
      if (xpp->getEventType() == xpp->TEXT)
        val = xpp->getText();
      validateSimpleContent(val, typeId, t);
      xpp->nextTag();
      if (xpp->getEventType() == xpp->END_TAG)
        {
          if (xpp->getName() != elemName)
            error("Syntax error");
        }

      else
        error("Expected a closing tag for " + elemName);
    }

  else
    {

      /*
        Perform Validation of Complex types
        Check for
        1.Is the tag name correct (this has to be right !!)
        2.Attributes ,if any should be valid
        3.Are there any mandatory (#required) attributes
        4. Call the routine recursively for its child elements
        5.Also check if the child elements follow the compositor
        rules viz ,all, sequence,choice
        6.Return the type container which has the
        correctly filled in values

      */
      ComplexType *ct = (ComplexType *) sParser->getType(typeId);
      if (ct->getNumAttributes() > 0)
        {
          for (int i = 0; i < ct->getNumAttributes(); i++)
            {
              Attribute at = ct->getAttribute(i);
              
              /*
                Check for the correctness of each attribute
              */
              string attVal = xpp->getAttributeValue("", at.getName());
              if (attVal.empty())
                {
                  if (at.isRequired())
                    error("Required attribute " + at.getName() +
                          " missing or empty");

                  else
                    continue;
                }
              TypeContainer *atCnt =
                t->getAttributeContainer(at.getName(), true);
              validateSimpleContent(attVal, at.getType(), atCnt);
            }
        }
      if (ct->getContentModel() == ct->SIMPLE)
        {
          xpp->nextToken();
          if (xpp->getEventType() == xpp->TEXT)
            {
              string val = xpp->getText();
              validateSimpleContent(val, ct->getContentType(), t);
            }
          xpp->nextTag();
          if (xpp->getEventType() == xpp->END_TAG)
            {
              if (xpp->getName() != elemName)
                error("Syntax error");
            }
          else
            error("Expected a closing tag for " + elemName);
        }
      if (ct->getNumChildren() > 0)
        {
          int gId = ct->getTopLevelGroup();
          list < Element > & pElemList = ct->pElemList();
          list < Element >::iterator pElem = pElemList.begin();
          ct->resetCounters();
          switch (gId)
            {
            case ComplexType::ALL:
              do
                {
                  xpp->nextTag();
                  if (xpp->getEventType() == xpp->END_TAG)
                    {
                      if (xpp->getName() == elemName)
                        break;
                      while (xpp->getEventType() != xpp->START_TAG)
                        xpp->nextTag();
                    }
                  if(!findElement(pElemList,pElem, xpp->getName()))
                    error("Could not find element " +xpp->getName()+" in "+elemName);
                  if (pElem->nOccurrences > 1)
                    error
                      ("Inside <all> an element can occur only once");
                  else
                    pElem->nOccurrences++;

                  validate(xpp, pElem->getType(),
                           t->getChildContainer(pElem->getName(), true));
                }
              while (true);

              /*
                If any element has minOccurs =1 then check if it was parsed
              */
              if (!ct->checkOccurrences())
                error("One or more mandatory elements is missing");
              break;
            case ComplexType::SEQ:

              do
                {
                  xpp->nextTag();
                  if (xpp->getEventType() == xpp->END_TAG)
                    {
                      if (xpp->getName() == elemName)
                        break;
                      while (xpp->getEventType() != xpp->START_TAG)
                        xpp->nextTag();
                    }

                  /* rewind =false because all elements occur in a sequence
                   */
                  if(!findElement(pElemList, pElem, xpp->getName(), false))
                    error("Could not find element " +xpp->getName()+" in "+elemName);
                  pElem->nOccurrences++;
                  if (pElem->nOccurrences > pElem->max())
                    error(pElem->getName() +
                          " Occurred more than the maximum permitted times");
                  validate(xpp, pElem->getType(),
                           t->getChildContainer(pElem->getName(), true));
                }
              while (true);

              /*
                If any element has minOccurs =1 then check if it was parsed
              */
              if (!ct->checkOccurrences())
                error("One or more mandatory elements is missing");
              break;
            case ComplexType::CHOICE:
              xpp->nextTag();
              findElement(pElemList, pElem , xpp->getName(), false);
              pElem->nOccurrences++;
              if (pElem->nOccurrences > pElem->max())
                error(pElem->getName() +
                      " Occurred more than the maximum permitted times");
              validate(xpp, pElem->getType(),
                       t->getChildContainer(pElem->getName(), true));
              xpp->nextTag();

              /*
                Only one of the choices is allowed
              */
              if (!
                  (xpp->getEventType() == xpp->END_TAG
                   && xpp->getName() == ct->getName()))
                error("Incorrect choice " + pElem->getName());

              /*
                If any element has minOccurs =1 then check if it was parsed
              */
              if (!ct->checkOccurrences())
                error("One or more mandatory elements is missing");
              break;
            }

          /*
            reset occurence ocunters
          */
          ct->resetCounters();
        }
    }
  return t;
}


TypeContainer * 
SchemaValidator::validate(string & val , int typeId,
            TypeContainer * ipTc )
{

      return    validateSimpleContent(val, typeId, ipTc);
}
/*
  This method validates all supported simple types
 */
TypeContainer * SchemaValidator::validateSimpleContent(string & val,
int sType,
TypeContainer *
ipTc)
{
    if (!ipTc)
        ipTc = new TypeContainer(sType, sParser);
    if (sParser->isBasicType(sType))
        extractSimpleType(val, sType, ipTc, 0);

    else
    {
        const XSDType *pType = sParser->getType(sType);
        SimpleType *st = (SimpleType *) pType;

//Its  a schema defined type
        int baseType = st->getBaseType();
        extractSimpleType(val, baseType, ipTc, st);
    }
    return ipTc;
}


void SchemaValidator::extractSimpleType(string & val, int basetype,
                                        TypeContainer * ipTc,
                                        SimpleType * st)
{
  istringstream istr(val);
  int x;
  double db;
  long l;
  char c;
  unsigned long ul;
  float f;
  bool b;
  Qname *q = 0;
  switch (basetype)
    {
    case XSDType::INTEGER:
    case XSDType::INT:
      istr >> x;

      //cout<<"Value extracted is "<<x;
      if (!st)
        {
          ipTc->setValue(x);
          break;
        }
      if (st->isValidInt(x))
        ipTc->setValue(x);

      else
        error("Invalid  int");
      break;
    case XSDType::BYTE:
      istr >> c;
      ipTc->setValue(c);
      break;
    case XSDType::FLOAT:
      istr >> f;

      //cout<<"Value extracted is "<<f;
      if (!st)
        {
          ipTc->setValue(f);
          break;
        }
      if (st->isValidFloat(f))
        ipTc->setValue(f);

      else
        error("Invalid  float");
      break;
    case XSDType::DOUBLE:
    case XSDType::DECIMAL:
      istr >> db;

      //cout<<"Value extracted is "<<db;
      ipTc->setValue(db);
      break;
    case XSDType::LONG:
      istr >> l;

      //cout<<"Value extracted is "<<l;
      ipTc->setValue(l);
      break;
    case XSDType::POSINT:
    case XSDType::ULONG:
      istr >> ul;

      //cout<<"Value extracted is "<<ul;
      ipTc->setValue(ul);
      break;
    case XSDType::BOOLEAN:
      istr >> b;

      //cout<<"Value extracted is "<<b;
      ipTc->setValue(b);
      break;
    case XSDType::QNAME:
      q = new Qname(val);
      ipTc->setValue(*q);
      break;
    case XSDType::STRING:
      if (!st)
        {
          ipTc->setValue(val);
          break;
        }
      if (st->isValidString(val))
        ipTc->setValue(val);

      else
        error("Invalid  string");
      break;
    case XSDType::ANYTYPE:
    case XSDType::ANYURI:
    case XSDType::NMTOKEN:
    case XSDType::NMTOKENS:
    case XSDType::TOKEN:
    case XSDType::NCNAME:
    case XSDType::DATE:
    case XSDType::DATETIME:
    case XSDType::TIME:
    default: //all these types are treated as strings
      ipTc->setValue(val);
      break;
    }
}

/*
 * This function searches for a child element in a complex type 
 * The iterator pElem is set to point to the found element
 * rewind controls whether a search is to be done by resetting pElem 
 * to the begining of the list 
 * Returns true if element is found else false 
 */

bool 
SchemaValidator::findElement(list < Element > &pElemList,
                                      list < Element >::iterator & pElem,
                                      string name, bool rewind)
{
     list < Element >::iterator pEnd=pElemList.end();
     list < Element >::iterator ptr=pElem;
    while (pElem != pEnd)
    {
        if (pElem->getName() == name)
            return true;
        else
            pElem++;
    }
    
    if(rewind)
    {
        pElem = pElemList.begin();
        while (pElem != ptr)
        {
            if (pElem->getName() == name)
                return true ;

            else
                pElem++;
        }
    }
    return false;
}


void SchemaValidator::error(string mesg)
{
    SchemaParserException spe(mesg + "\nError in SchemaValidator\n");
    throw spe;
}


//TODO validation of <any>
//TODO validation of base64binary,hexBinary
//TODO validation of types derived by extension/restriction of complex types
