/* 
 * wsdlpull - A C++ parser  for WSDL  (Web services description language)
 * Copyright (C) 2003 Vivek Krishna
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

/*
 * This class stores the list of parsed types 
 */
 
#include <schemaparser/TypesTable.h>

TypesTable::TypesTable()
{
    currentId = XSDType::ANYURI + 1;
    numTypes = 0;
    nExtRefs = 0;
    typesArray = new XSDType *[nSize = 10];

//map of simple types
    basicTypes["string"] = XSDType::STRING;
    basicTypes["integer"] = XSDType::INTEGER;
    basicTypes["int"] = XSDType::INT;
    basicTypes["byte"] = XSDType::BYTE;
    basicTypes["positiveInteger"] = XSDType::POSINT;
    basicTypes["unsignedInt"] = XSDType::UINT;
    basicTypes["long"] = XSDType::LONG;
    basicTypes["unsignedLong"] = XSDType::ULONG;
    basicTypes["short"] = XSDType::SHORT;
    basicTypes["unsignedShort"] = XSDType::USHORT;
    basicTypes["decimal"] = XSDType::DECIMAL;
    basicTypes["float"] = XSDType::FLOAT;
    basicTypes["double"] = XSDType::DOUBLE;
    basicTypes["boolean"] = XSDType::BOOLEAN;
    basicTypes["time"] = XSDType::TIME;
    basicTypes["dateTime"] = XSDType::DATETIME;
    basicTypes["date"] = XSDType::DATE;
    basicTypes["token"] = XSDType::TOKEN;
    basicTypes["QName"] = XSDType::QNAME;
    basicTypes["NCName"] = XSDType::NCNAME;
    basicTypes["NMTOKEN"] = XSDType::NMTOKEN;
    basicTypes["NMTOKENS"] = XSDType::NMTOKENS;
    basicTypes["base64Binary"] = XSDType::BASE64BIN;
    basicTypes["hexBinary"] = XSDType::HEXBIN;
    basicTypes["anyType"] = XSDType::ANYTYPE;
    basicTypes["any"] = XSDType::ANY;
    basicTypes["anyURI"] = XSDType::ANYURI;
}


TypesTable::~TypesTable()
{
    clean();
}


void
TypesTable::clean()
{
    for (map < string, int >::iterator it = Id.begin(); it != Id.end();
        ++it)
    delete getTypePtr(it->second);
    numTypes = 0;
    if (typesArray)
    {
        delete[] typesArray;
        typesArray = 0;
    }
}


int TypesTable::addType(XSDType * type)
{
    Qname qn = type->getQname();
    string type_name(qn.getLocalName());
    int i = 0;

//  string ns(qn.getNamespace());
    if (type_name.empty())
    {

//create an anonymous type name
        ostringstream tmp_name_str;
        tmp_name_str << "type" << numTypes;
        type_name = tmp_name_str.str();
        type->setName(type_name);
        type->setAnonymous(true);                 //auto generated name
    }
    ensureCapacity();

//add the typename and its id  to the map
    if ((i = Id[type_name]) != 0)
    {

//this type was refernced earlier.
        typesArray[i - (XSDType::ANYURI + 1)] = type;
        type->setTypeId(i);
        return i;
    }

    else
    {
        Id[type_name] = currentId;
        type->setTypeId(currentId);
        typesArray[numTypes] = type;
        currentId++;
        numTypes++;
        return currentId - 1;
    }
}




//get the type id of a type
int TypesTable::getTypeId(Qname & name, bool create)
{
  int typeId;
  if (name.getNamespace() == SchemaUri)
    {

      //This is one of the basic types
      typeId = basicTypes[name.getLocalName()];
      if(typeId==0) //if this is a basic type which is not mapped,treat as string
        typeId=XSDType::STRING;
    }

  else if (name.getNamespace() == m_tnsUri)
    typeId = Id[name.getLocalName()];

  else
    return 0;                                 //the Type does not belong to this schema
  if (typeId == 0 && create)
    {

      //handle forward reference
      //create an id and return its value
      Id[name.getLocalName()] = currentId;
      ensureCapacity();
      typesArray[numTypes] = 0;
      currentId++;
      numTypes++;
      typeId = currentId - 1;
    }
  return typeId;
}


int TypesTable::addExternalTypeId(Qname & type, XSDType * pType)
{
  for (int i = 0; i < nExtRefs; i++)
    if (m_extRefs[i].qname == type)
      return m_extRefs[i].localTypeId;
  m_extRefs[nExtRefs].qname = (pType)?pType->getQname():type;
  ensureCapacity();
  m_extRefs[nExtRefs].localTypeId = currentId;
  typesArray[numTypes] = pType;
  currentId++;
  numTypes++;
  return m_extRefs[nExtRefs++].localTypeId;

}

//adds a type into a type table for a given type id
//used for types present in imported schemas but referenced in current schema
int TypesTable::addExtType(XSDType * type, int localId)
{
  int index = localId - XSDType::ANYURI - 1;
  if (index >= numTypes)
    return 0;
  typesArray[index] = type;
  return localId;
}

//increase the array storage if necessary
void TypesTable::ensureCapacity()
{
  if (numTypes >= nSize)
    {
      XSDType **tempArray = new XSDType *[numTypes + 5];
      for (int ind = 0; ind < nSize; ind++)
        tempArray[ind] = typesArray[ind];
      delete typesArray;
      typesArray = tempArray;
      nSize = numTypes + 5;
    }
}


/*
  Takes a XSD type  and an accessor name to determine if the
  accessor is a descendant of the given type
  If found ,returns the schematype Id of the found element
  and populates an array of indices denoting the path
  from the parent to accessor .
  otherwise return 0(invalid type)

*/
int TypesTable::getCompleteXpath(int elemId, string & childName,
int *xPath, int limits, int &offset)
{
    int childId = 0, type = 0, i = 0, childIndex;
    bool found = false;
    if (xPath == 0 || limits == 0)
        return 0;
    XSDType *pType = getTypePtr(elemId);
    if (pType == 0)
        return 0;
    if (pType->isSimple())
        return 0;
    ComplexType *ct = (ComplexType *) pType;
    for (i = 0; i < ct->getNumChildren(); i++)
        if (ct->getChildName(i) == childName)
    {
        childId = ct->getChildType(i);
        childIndex = i;
    }
    if (childId == 0)
    {

/*
 childName is not a child of elemId,
 so call this method recursively
*/
        for (int i = 0; i < ct->getNumChildren() && !childId; i++)
        {
            if (childId =
                getCompleteXpath(ct->getChildType(i), childName, xPath + 1,
                limits - 1, ++offset))
                xPath[0] = i;

            else
                offset--;
        }
    }

    else
    {
        xPath[0] = childIndex;
        offset++;
    }
    return childId;
}


bool TypesTable::detectUndefinedTypes(void)
{
    for (int i = 0; i < numTypes; i++)
        if (typesArray[i] == 0)
            return true;
    return false;
}


void TypesTable::resolveForwardElementRefs(const string & name, Element & e)
{
    for (int i = 0; i < numTypes; i++)
        if (typesArray[i] != 0)
    {
        if (!typesArray[i]->isSimple())
        {
            ComplexType *ct = (ComplexType *) typesArray[i];
            ct->matchElementRef(name, e);
        }
    }
}


void TypesTable::resolveForwardAttributeRefs(const string & name, Attribute & a)
{
    for (int i = 0; i < numTypes; i++)
        if (typesArray[i] != 0)
    {
        if (!typesArray[i]->isSimple())
        {
            ComplexType *ct = (ComplexType *) typesArray[i];
            ct->matchAttributeRef(name, a);
        }
    }
}


//for debug purposes
void TypesTable::printUndefinedTypes(ostream & out)
{
    for (map < string, int >::iterator it = Id.begin(); it != Id.end();
        ++it)
    {
        if (getTypePtr(it->second) == 0)
            out << m_tnsUri << ":" << it->first << dbsp;
    }
    out << endl;
}

