/* 
 * wsdlpull - A C++ parser for WSDL (Web services description
 * language) Copyright (C) 2003 Vivek Krishna
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *
 */

#include <wsdlparser/WsdlParser.h>

WsdlParser::WsdlParser(istream & in, ostream & out)
           :m_ErrorOccured(false), 
           istr(in),
           ostr(out)
{
    xParser= new XmlPullParser(in);
    xParser->setFeature(FEATURE_PROCESS_NAMESPACES, true);
    xParser->require(xParser->START_DOCUMENT, "", "");
    m_State = m_Element = START;
    m_messages_list.clear();
    m_bindings_list.clear();
    m_porttypes_list.clear();
    nMessage = nOps = nPortTypes = nBindings = nImports = 0;
    m_Doc=0;
    int i = 0;
    for (i = 0; i < MAX_EXT; i++)
    {
        m_wsdlExtensions[i].we = 0;
        m_wsdlExtensions[i].spe = 0;
    }
    nExtensions = 0;
    for (i = 0; i < MAX_SCHEMAS; i++)
    {
        m_schemaParser[i] = 0;
    }
    m_schemaParser[0] = new SchemaParser(xParser, m_tnsUri);
    nSchemas = 0;                                 //create a dummy schema parser,for parsing Wsdls with no <types> tag
}


WsdlParser::~WsdlParser()
{
    int i = 0;
    for (list < const Message * >::iterator pMessage =
        m_messages_list.begin(); pMessage != m_messages_list.end();
        pMessage++)
    delete(*pMessage);
    for (list < const Binding * >::iterator pBinding =
        m_bindings_list.begin(); pBinding != m_bindings_list.end();
        pBinding++)
    delete(*pBinding);
    for (list < const PortType * >::iterator pPortType =
        m_porttypes_list.begin(); pPortType != m_porttypes_list.end();
        pPortType++)
    delete(*pPortType);
    if(nSchemas){
    for (i = 0; i <  nSchemas; i++)
        delete m_schemaParser[i];
    }  else
	delete m_schemaParser[0];
	    
    for (i = 0; i < nOps; i++)
        delete m_Ops[i];

    for (list < string * >::iterator pString =
        m_docs_list.begin(); pString != m_docs_list.end();
        pString++)
    delete(*pString);

    delete xParser;

}


//public  APIs
int  WsdlParser::getNextElement()
{
    next();
    return m_Element;
}


const string*  WsdlParser::getDocumentation() 
{
    return  (const string*) m_Doc;
}


string  WsdlParser::getTargetNamespace(void)
{
    return m_tnsUri;
}


string  WsdlParser::getName()
{
    return m_name;
}


void  WsdlParser::addExtensibilityHandler(WsdlExtension * ext)
{
    m_wsdlExtensions[nExtensions++].we = ext;
} 


int  WsdlParser::getEventType()
{
    return m_Element;
}


const Binding * WsdlParser::getBinding()
{
    if (m_Element != BINDING)
    {
        error ("Attempted to extract a Binding when ,no binding was parsed",1);
    return 0;
    }
    else
    {
            return m_bindings_list.back();
    }
}


const Binding * WsdlParser::getBinding(Qname & qn)
{
    string name = qn.getLocalName();
    if (!qn.getPrefix().empty())
        qn.setNamespace(getNamespace(qn.getPrefix()));

    else
        qn.setNamespace(m_tnsUri);
    if (m_tnsUri != qn.getNamespace())
        return 0;
    for (list < const Binding * >::iterator pBinding =
        m_bindings_list.begin(); pBinding != m_bindings_list.end();
        pBinding++)
    if ((*pBinding)->getName() == name)
        return *pBinding;
    return 0;
}


list < const Binding *>* WsdlParser::getBindings()
{
    return &(m_bindings_list);
}


const PortType * WsdlParser::getPortType()
{
    if (m_Element != PORT_TYPE)
    {
        error ("Attempted to extract a PortType when ,no PortType was parsed",1);
        return 0;
    }
    else
    {
            return m_porttypes_list.back();
    }
}


const PortType * WsdlParser::getPortType(Qname & qn)
{
    string name = qn.getLocalName();
    if (!qn.getPrefix().empty())
        qn.setNamespace(getNamespace(qn.getPrefix()));

    else
        qn.setNamespace(m_tnsUri);
    if (m_tnsUri != qn.getNamespace())
        return 0;
    for (list < const PortType * >::iterator pPortType =
        m_porttypes_list.begin(); pPortType != m_porttypes_list.end();
        pPortType++)
    if ((*pPortType)->getName() == name)
        return *pPortType;
    return 0;
}


list < const PortType *>* WsdlParser::getPortTypes()
{
    return &m_porttypes_list;
}


list < const Operation *>* WsdlParser::getOperations(Qname & portType)
{
    const PortType *pt = getPortType(portType);
    int num = pt->getNumOps();
    list < const Operation *>*plOps = 0;
    if (num > 0)
    {
        plOps = new list < const Operation *>();
        const Operation *op;
        for (int i = 0; i < num; i++)
        {
            op = pt->getOperation(i);
            plOps->push_back(op);
        }
    }
    return plOps;
}


const Operation *
WsdlParser::getOperation(Qname & portType, Qname & q)
{
    const PortType *pt = getPortType(portType);
    int num = pt->getNumOps();
    if (num > 0)
    {
        const Operation *op;
        for (int i = 0; i < num; i++)
        {
            op = pt->getOperation(i);
            if (op->getName() == q.getLocalName())
                return op;
        }
    }
    return 0;
}


const Message * WsdlParser::getMessage()
{
    if (m_Element != MESSAGE)
    {
        error ("Attempted to extract a Message when ,no Message was parsed",1);
	return 0;
    }
    else
    {
            return m_messages_list.back();
    }
}


const Message * WsdlParser::pgetMessage(Qname qn)
{
    return getMessage(qn);
}


const Message * WsdlParser::getMessage(Qname & qn)
{
    string name = qn.getLocalName();
    if (!qn.getPrefix().empty())
        qn.setNamespace(getNamespace(qn.getPrefix()));

    else
        qn.setNamespace(m_tnsUri);
    if (m_tnsUri != qn.getNamespace())
        return 0;
    for (list < const Message * >::iterator pMessage =
        m_messages_list.begin(); pMessage != m_messages_list.end();
        pMessage++)
    if ((*pMessage)->getName() == name)
        return *pMessage;
    return 0;
}


list < const Message *>* WsdlParser::getMessages()
{
    return &m_messages_list;
}


const SchemaParser ** WsdlParser::getSchemaParsers() const
{
    return (const SchemaParser **) m_schemaParser;
} 

int  WsdlParser::getNumSchemas() const
{
    return nSchemas;
}


const SchemaParser * WsdlParser::getSchemaParser(string targetNamespace) const
{
    if (targetNamespace == SchemaUri)
        return 0;
    for (int i = 0; i < nSchemas; i++)
        if (m_schemaParser[i]->getTargetNamespace() == targetNamespace)
            return (const SchemaParser *) m_schemaParser[i];
    return 0;
}


const SchemaParser * WsdlParser::getSchemaParser(int schemaId) const
{
    return (const SchemaParser *) m_schemaParser[schemaId];
}


////////// private methods
inline bool  isValidWsdlElement(int id)
{
    if (id >= 0)
        return true;

    else
        return false;
}


int  WsdlParser::peek(bool lookahead)
{

//event Type returned by XML pull parser
    int event_type, tmp_event_type = xParser->getEventType();
    int tmpState = m_State;
    if (m_State == END)
        return m_State;

    do
    {
        if (lookahead == true || m_State == START || m_State == NONE)
            xParser->nextTag();

        else
            return m_State;
        event_type = xParser->getEventType();
        string tag = xParser->getName();
        switch (event_type)
        {
            case xParser->START_DOCUMENT:
                if (m_State != START)
                    error("Syntax error at the start");
                break;
            case xParser->START_TAG:
                if (xParser->getNamespace() != wsdlUri
                    &&xParser->getNamespace() != SchemaUri)
                    m_State = EXTENSIBILITY;

                else if (tag == "definitions")
                    m_State = DEFINITION;

                else if (tag == "documentation")
                    m_State = DOCUMENTATION;

                else if (tag == "annotation")
                    m_State = ANNOTATION;

                else if (tag == "import")
                    m_State = IMPORT;

                else if (tag == "schema")
                    m_State = SCHEMA;

                else if (tag == "types")
                    m_State = TYPES;

                else if (tag == "message")
                    m_State = MESSAGE;

                else if (tag == "port")
                    m_State = PORT;

                else if (tag == "operation")
                    m_State = OPERATION;

                else if (tag == "portType")
                    m_State = PORT_TYPE;

                else if (tag == "input")
                    m_State = INPUT;

                else if (tag == "output")
                    m_State = OUTPUT;

                else if (tag == "fault")
                    m_State = FAULT;

                else if (tag == "part")
                    m_State = PART;

                else if (tag == "binding")
                    m_State = BINDING;

                else if (tag == "service")
                    m_State = SERVICE;

                else
                    error("Unknown Tag " + tag);
                break;
            case xParser->END_TAG:
                if (tag == "definitions")
                    m_State = END;

                else
                {

/*
  If its one of the 7 Wsdl elements
  set the State to NONE
  */
                    if (tag == "types" || tag == "message"
                        || tag == "portType"  ||tag == "import"
                        || (tag == "binding"  &&m_State !=
                        EXTENSIBILITY)  ||tag == "service")
                        return m_State = NONE;

                    else
                        return peek(lookahead);   //get the next tag
                }
                break;
            case xParser->TEXT:
            case xParser->ENTITY_REF:
            case xParser->COMMENT:
            case xParser->PROCESSING_INSTRUCTION:
            case xParser->CDSECT:
                xParser->getText();
                break;
            case xParser->DOCDECL:
                error("Doc Declaration ??");
                break;
            default:
                error("Unknown Wsdl tag");
                break;
        }
    } while (event_type != xParser->END_DOCUMENT
        && tmpState == m_State  &&event_type ==
        tmp_event_type);
    return m_State;
}


//this method looks at the top level Wsdl elements
int  WsdlParser::next()
{
    try
    {
        switch (peek(false))
        {
            case START:
                m_Element = START;
                break;
            case DEFINITION:
                parseDefinitions();
                peek();
                m_Element = DEFINITION;
                break;
            case DOCUMENTATION:
                m_Doc=parseDoc();
                m_Element = DOCUMENTATION;
                break;
            case ANNOTATION:
                parseAnnotation();
                m_Element = ANNOTATION;
                break;
            case IMPORT:
                parseImport();
                m_Element = IMPORT;
                break;
            case TYPES:
                parseTypes();
                m_Element = TYPES;
                break;
            case MESSAGE:
                parseMessage();
                m_Element = MESSAGE;
                break;
            case PORT_TYPE:
                parsePortType();
                m_Element = PORT_TYPE;
                break;
            case EXTENSIBILITY:
                handleExtensibilityElement(DEFINITION);
                peek();
                m_Element = EXTENSIBILITY;
                break;
            case SERVICE:
                parseService();
                m_Element = SERVICE;
                break;
            case BINDING:
                parseBinding();
                m_Element = BINDING;
                break;
            case END:
                m_Element = END;
                return m_State;
            default:
                error("Syntax error");
        }
        return m_State;
    }
    catch(WsdlException we)
    {
        we.line = xParser->getLineNumber();
        we.col = xParser->getColumnNumber();
        m_ErrorOccured = true;
        m_Element = END;
        cout << we.description << " at " << we.line << "," << we.
            col << endl;
        return m_State = END;
    }
    catch(XmlPullParserException xe)
    {
        cout << xe.description << endl;
        m_ErrorOccured = true;
        m_Element = END;
        return m_State = END;
    }
}


/*
Parse a documentation tag
*/
string* 
WsdlParser::parseDoc()
{
    int  event_type;
    string*  documentation = new string();
    if (m_State != DOCUMENTATION)
        error("syntax error");

    do
    {
        xParser->nextToken();
        if (xParser->getEventType() == xParser->TEXT)
            *documentation += xParser->getText();
        if (xParser->getEventType() == xParser->END_TAG
            &&xParser->getName() == "documentation")
            break;
    } while (true);
    m_docs_list.push_back(documentation);
    return documentation;
}


/*
Parse Annotation
*/
void  WsdlParser::parseAnnotation()
{
    int event_type;
    if (m_State != ANNOTATION)
        error("syntax error");

    do
    {
        xParser->nextToken();
        if (xParser->getEventType() == xParser->END_TAG
            &&xParser->getName() == "annotation")
            break;
    } while (true);
}


/*Parses the definition tag
  If any extensibility namespaces are defined then the relevant
  information is stored
*/
void  WsdlParser::parseDefinitions()
{
    if (m_State != DEFINITION)
        error("syntax error");
    m_tnsUri = xParser->getAttributeValue("", "targetNamespace");
    int i = 0;
    for (i = xParser->getNamespaceCount(xParser->getDepth()) - 1;
        i > xParser->getNamespaceCount(xParser->getDepth() - 1) - 1; i--)
    {
        if (xParser->getNamespaceUri(i) == m_tnsUri)
            m_tnsPrefix = xParser->getNamespacePrefix(i);

/* 
 * Associate the extension prefixes with the handlers.
 * It is asssumed that by this time all the extensibility handlers have been registered .
 * Check if the namespace defined here matches that Uri ,whose namespace the handler handles .
 */
        for (int j = 0; j < MAX_EXT; j++)
            if (m_wsdlExtensions[j].we != 0
            &&xParser->getNamespaceUri(i) ==
            m_wsdlExtensions[j].we->getNamespace())
        {
            m_wsdlExtensions[j].we->setNamespacePrefix(xParser->
                getNamespacePrefix
                (i));
            m_wsdlExtensions[j].we->setStartId(MAX_EXT_XML * j + 1);

/*
 * If there is a schema associated with the extensibility namespace
 * use the schema parser to parse its types.
 */
            XmlPullParser * xtmpParser =
                m_wsdlExtensions[j].we->getExtensibilitySchema();
            if (xtmpParser == 0)
                continue;
            SchemaParser * xtmpSchemaParser =
                new SchemaParser(xtmpParser,
                m_wsdlExtensions [j].we->getNamespace());
            if (xtmpSchemaParser->parseSchemaTag())
            {
                m_wsdlExtensions[j].spe = xtmpSchemaParser;
                m_wsdlExtensions[j].we->
                    setSchemaParser(xtmpSchemaParser);
                m_wsdlExtensions[j].we->setWsdlParser(this);
            }

            else
                error("Error parsing extensibility schema for " +
                    m_wsdlExtensions[j].we->getNamespace());
            delete xtmpParser;
        }
    }
    int num_attr = xParser->getAttributeCount();
    if (num_attr < 0)
        error("Atleast a targetNamespace attribute is needed");
    for (i = 0; i < num_attr; i++)
    {
        if (xParser->getAttributeName(i) == "name")
        {
            m_name = xParser->getAttributeValue(i);
            continue;
        }

        else if (xParser->getAttributeName(i) != "targetNamespace")
        {                                         //this is to handle extensibility attributes
            handleExtensibilityAttributes(xParser->getAttributePrefix(i),
                xParser->getAttributeName(i));
        }
    }
    return;
}


void  WsdlParser::parseImport()
{
    if (m_State != IMPORT)
        error("syntax error");
    m_imports[nImports].ns = xParser->getAttributeValue("", "namespace");
    m_imports[nImports].loc = xParser->getAttributeValue("", "location");
    nImports++;
    peek();
}


void  WsdlParser::parseMessage()
{
  if (m_State != MESSAGE)
    error("syntax error");
  Message * m = new Message();
  int num_att = xParser->getAttributeCount();
  for (int i = 0; i < num_att; i++)
    {
      if ("name" == xParser->getAttributeName(i) &&
          (xParser->getAttributePrefix(i)).empty())
        m->setName(xParser->getAttributeValue(i));

      else if (!(xParser->getAttributePrefix(i)).empty())
        m->addExtAtts(handleExtensibilityAttributes
                      (xParser->getAttributePrefix(i),
                       xParser->getAttributeName(i)));

      else                                      //TODO  Wsdl1.1 ? 1.2
        error("Syntax error..unrecognized attribute");
    }
  if (m->getName() == "")
    error("syntax error <message> name required");
  peek();
  try
    {
      if (m_State == DOCUMENTATION)
        {
          m->setDocumentation(parseDoc());
          peek();
        }

      //parse all the parts in the message
      //TODO .if a part has a type reference ,check that only one part is allowed in the message
      if (m_State == PART)
        {
          while (m_State == PART)
            {
              string p_name;
              int type_id = 0, schemaId = -1;
              refType reftype = (refType) None;
              int num_att = xParser->getAttributeCount();
              int p_extId = 0;
              for (int i = 0; i < num_att; i++)
                {
                  if ("name" == xParser->getAttributeName(i) &&
                      //Wsdl attribute name must have a null prefix
                      (xParser->getAttributePrefix(i)).empty())
                    p_name = xParser->getAttributeValue(i);

                  else if (("type" == xParser->getAttributeName(i)
                            &&xParser->getAttributePrefix(i).empty())
                           ||("element" == xParser->getAttributeName(i)
                              &&xParser->getAttributePrefix(i).empty()))
                    {
                      if (reftype != None)
                        error
                          ("either type or element must occur(only once) in part ");
                      if ("type" == xParser->getAttributeName(i))
                        reftype = Type;

                      else
                        reftype = Elem;
                      Qname type(xParser->getAttributeValue(i));
                      type.setNamespace(getNamespace(type.getPrefix()));
                      if (reftype == Type)
                        {

                          //get the type id
                          type_id = getTypeId(type);
                          if (type_id == 0)
                            error("Could not resolve type " +
                                  type.getNamespace() + ":" +
                                  type.getLocalName());
                        }

                      else
                        {

                          //get the element id
                          type_id = getElementId(type);
                          if (type_id == -1)
                            error("Could not resolve element " +
                                  type.getNamespace() + ":" +
                                  type.getLocalName());
                        }

                      //if the ref type is "element",the id is that of a global element and not a type
                      //get the schema parser of the namespace to which "type" belongs
                      schemaId = getSchema(type);
                    }

                  else if (!(xParser->getAttributePrefix(i)).empty())
                    p_extId = handleExtensibilityAttributes(xParser->
                                                            getAttributePrefix
                                                            (i),
                                                            xParser->

                                                            getAttributeName
                                                            (i));

                  else
                    error("Syntax error");
                }
              peek();
              if (m_State == DOCUMENTATION)
                {
                  parseDoc();
                  peek();
                }
              m->addPart(p_name, reftype, type_id, schemaId, p_extId);
            }
        }
    }
  catch(WsdlException we)
    {
      we.line = xParser->getLineNumber();
      we.col = xParser->getColumnNumber();
      throw we;
    }

  //now parse the extensibility elements
  if (m_State == EXTENSIBILITY)
    {
      while (m_State == EXTENSIBILITY)
        {
          m->addExtElement(handleExtensibilityElement(MESSAGE));
          peek();
        }
    }

  //add the message to the list
  putMessage(m);
  return;
}


//This is a key function ..parse port types
PortType * WsdlParser::parsePortType()
{
    if (m_State != PORT_TYPE)
        return 0;
    if (nPortTypes >= MAX_PTYPES)
        error("max port types limit reached");
    PortType * pt = new PortType();
    int num_att = xParser->getAttributeCount();
    for (int i = 0; i < num_att; i++)
    {
        if ("name" == xParser->getAttributeName(i) &&
//Wsdl attribute name must have a null prefix
            (xParser->getAttributePrefix(i)).empty())
            pt->setName(xParser->getAttributeValue(i));

        else if (!(xParser->getAttributePrefix(i)).empty())
            pt->addExtAtts(handleExtensibilityAttributes
                    (xParser->getAttributePrefix(i),
                    xParser->getAttributeName(i)));

        else
            error("Syntax error..unrecognized attribute");
    }
    if (pt->getName() == "")
        error("syntax error <PortType> name required");
    peek();
    if (m_State == DOCUMENTATION)
    {
        pt->setDocumentation(parseDoc());
        peek();
    }
    if (m_State == OPERATION)
    {

//parse all the operations in the port type
        while (m_State == OPERATION)
        {
            Operation * op = parseOperation();
            pt->addOp(op);
        }
        if (m_State == EXTENSIBILITY)
        {

//now parse the extensibility elements
            while (m_State == EXTENSIBILITY)
            {
                pt->addExtElement(handleExtensibilityElement(PORT_TYPE));
                peek();
            }
        }
    }
    putPortType(pt);
    return pt;
}


//Returns an operation element
Operation * WsdlParser::parseOperation()
{
    Operation * op = new Operation();
    if (m_State != OPERATION)
        error("syntax error");
    if (nOps >= MAX_OPS)
        error("Max limit for number of operations  reached");
    int num_att = xParser->getAttributeCount();
    for (int i = 0; i < num_att; i++)
    {
        if ("name" == xParser->getAttributeName(i) &&
            (xParser->getAttributePrefix(i)).empty())
            op->setName(xParser->getAttributeValue(i));

//Wsdl attribute name must have a null prefix

        else if (!(xParser->getAttributePrefix(i)).empty())
            op->addExtAtts(handleExtensibilityAttributes
                    (xParser->getAttributePrefix(i),
                    xParser->getAttributeName(i)));

        else if ("parameterOrder" == xParser->getAttributeName(i))
        {
        }

        else
            error("Syntax error..unrecognized attribute");
    }
    if (op->getName() == "")
        error("syntax error <operation> name required");
    peek();
    if (m_State == DOCUMENTATION)
    {
        op->setDocumentation(parseDoc());
        peek();
    }
    if (m_State == INPUT)
    {
        op->setMessage(pgetMessage
            (Qname(xParser->getAttributeValue("", "message"))),
            Input);
        peek();
        if (m_State == OUTPUT)
        {
            op->setMessage(pgetMessage
                (Qname
                (xParser->getAttributeValue("", "message"))),
                Output);
            peek();
        }
        while (m_State == FAULT)
        {
            op->setMessage(pgetMessage
                (Qname
                (xParser->getAttributeValue("", "message"))),
                Fault);
            peek();
        }
    }

    else if (m_State == OUTPUT)
    {
        op->setMessage(pgetMessage
            (Qname(xParser->getAttributeValue("", "message"))),
            Output);
        peek();
        if (m_State == INPUT)
        {
            op->setMessage(pgetMessage
                (Qname
                (xParser->getAttributeValue("", "message"))),
                Input);
            peek();
        }
        while (m_State == FAULT)
        {
            op->setMessage(pgetMessage
                (Qname
                (xParser->getAttributeValue("", "message"))),
                Fault);
            peek();
        }
    }
    if (m_State == DOCUMENTATION)
    {
        op->setDocumentation(parseDoc());
        peek();
    }
    if (m_State == EXTENSIBILITY)
        while (m_State == EXTENSIBILITY)
    {
        op->addExtElement(handleExtensibilityElement(OPERATION));
        peek();
    }

//op->setId (nOps);
    m_Ops[nOps++] = op;
    return op;
}


void  WsdlParser::parseTypes()
{
    peek();
    if (m_State == DOCUMENTATION)
    {
        parseDoc();
        peek();
    }
    try
    {
        while (m_State == SCHEMA)
        {
            if (nSchemas == 0)
                delete m_schemaParser[nSchemas];
            m_schemaParser[nSchemas] =
                new SchemaParser(xParser, m_tnsUri);
            if (!m_schemaParser[nSchemas]->parseSchemaTag())
                error("Error parsing schema types");
            peek();
            nSchemas++;
        }
        for (int i = 0; i < nSchemas; i++)
        {
            m_schemaParser[i]->addImports(m_schemaParser, nSchemas);
            if (!m_schemaParser[i]->finalize())
                error("Undefined types");
        }
    }
    catch(SchemaParserException spe)
    {
        WsdlException we(spe.description);
        we.col = spe.col;
        we.line = spe.line;
        we.WsdlState = m_State;
        throw we;
    }
}


void  WsdlParser::putMessage(Message * m)
{

//m->setId (nMessage++);
    m_messages_list.push_back(m);
} 

void  WsdlParser::putBinding(Binding * bn)
{
    nBindings++;
    m_bindings_list.push_back((const Binding *) bn);
} 

void  WsdlParser::putPortType(PortType * pt)
{
    nPortTypes++;
    m_porttypes_list.push_back((const PortType *) pt);
} 

int  WsdlParser::handleExtensibilityElement(int parent)
{
    WsdlExtension * we = getPtr(xParser->getNamespace());
    if (we == 0)
        return 0;

    else
        return we->handleElement(parent, xParser);
}


int
WsdlParser::handleExtensibilityAttributes(string prefix, string name)
{
    WsdlExtension * we = getPtr(getNamespace(prefix));
    if (we == 0)
        return 0;

    else
        return we->handleAttribute(m_State, name, xParser);
}


/*
 parameter Ns is the prefix or the namespace of the extensibility handler
*/
WsdlExtension * WsdlParser::getPtr(string Ns)
{
    for (int i = 0; i < MAX_EXT; i++)
        if (m_wsdlExtensions[i].we != 0 &&
        (m_wsdlExtensions[i].we->isNamespaceHandler(Ns)))
            return m_wsdlExtensions[i].we;
    return 0;
}


void  WsdlParser::parseBinding()
{
    Binding * bn = new Binding();
    const PortType *pt = 0;
    int opBinding, inputBinding, outputBinding, faultBinding, index,
        bindingInfo;
    opBinding = inputBinding = outputBinding = faultBinding = index =
        bindingInfo = 0;
    if (m_State != BINDING)
        error("syntax error");
    int num_att = xParser->getAttributeCount();
    int i;
    for (i = 0; i < num_att; i++)
    {
        if ("name" == xParser->getAttributeName(i) &&
            (xParser->getAttributePrefix(i)).empty())
            bn->setName(xParser->getAttributeValue(i));

        else if ("type" == xParser->getAttributeName(i) &&
            (xParser->getAttributePrefix(i)).empty())
        {
            Qname q(xParser->getAttributeValue(i));
            pt = getPortType(q);
            bn->setPortType(pt);
        }

        else
            error("Syntax error..unrecognized attribute");
    }
    peek();
    if (m_State == DOCUMENTATION)
    {
        bn->setDocumentation(parseDoc());
        peek();
    }
    if (m_State == EXTENSIBILITY)
    {
        while (m_State == EXTENSIBILITY)
        {
            bn->setBindingInfo(bindingInfo =
                handleExtensibilityElement(BINDING));
            peek();
        }
    }
    while (m_State == OPERATION)
    {
        num_att = xParser->getAttributeCount();
        const Operation *op;
        for (i = 0; i < num_att; i++)
        {
            if ("name" == xParser->getAttributeName(i) &&
                (xParser->getAttributePrefix(i)).empty())
            {
                Qname q(xParser->getAttributeValue(i));
                op = pt->getOperation(q);
            }

            else
                error("Unrecognized attribute");
        }
        index = bn->addOperation(op);
        peek();
        while (m_State == EXTENSIBILITY)
        {
            opBinding = handleExtensibilityElement(OPERATION);
            if(opBinding) bn->addOpBinding(index, opBinding);
            peek();
        }
        if (m_State == INPUT)
        {
            peek();
            while (m_State == EXTENSIBILITY)
            {
                inputBinding = handleExtensibilityElement(OPERATION);
                if(inputBinding) bn->addInputBinding(index, inputBinding);
                peek();
            }
        }
        if (m_State == OUTPUT)
        {
            peek();
            while (m_State == EXTENSIBILITY)
            {
                outputBinding = handleExtensibilityElement(OPERATION);
                if(outputBinding) bn->addOutputBinding(index, outputBinding);
                peek();
            }
        }
        if (m_State == FAULT)
        {
            peek();
            while (m_State == EXTENSIBILITY)
            {
                faultBinding = handleExtensibilityElement(OPERATION);
                peek();
                if(faultBinding) bn->addFaultBinding(index, faultBinding);
            }
        }
    }
        putBinding(bn);
}


void  WsdlParser::parseService()
{
    if (m_State != SERVICE)
        error("Syntax error");
    string serviceName;
    int num_att = xParser->getAttributeCount();
    int i;
    for (i = 0; i < num_att; i++)
    {
        if ("name" == xParser->getAttributeName(i) &&
            (xParser->getAttributePrefix(i)).empty())
            serviceName = xParser->getAttributeValue(i);

        else
            error("Unrecognized attribute");
    }
    peek();
    if (m_State == DOCUMENTATION)
    {
        parseDoc();
        peek();
    }
    while (m_State == PORT)
    {
        string str;
        Binding * bn = 0;;
        int serviceExtId = 0;
        num_att = xParser->getAttributeCount();
        for (i = 0; i < num_att; i++)
        {
            if ("binding" == xParser->getAttributeName(i) &&
                (xParser->getAttributePrefix(i)).empty())
                str = xParser->getAttributeValue(i);
        }
        Qname bindingName(str);
        bn = (Binding *) getBinding(bindingName);
        peek();
        if (m_State == DOCUMENTATION)
        {
            parseDoc();
            peek();
        }
        if (m_State == EXTENSIBILITY)
        {
            serviceExtId = handleExtensibilityElement(BINDING);
            peek();
        }
        if (bn != 0)
            bn->setServiceExtId(serviceExtId);
    }
}


/*
 * returns the id of the schema to which "type" belongs
 */
int  WsdlParser::getSchema(Qname & type)
{
    if (type.getNamespace() == SchemaUri)
        return 0;                                 //this is a primitve type ,simple instance of schemaparser will do.
    for (int i = 0; i < nSchemas; i++)
        if (m_schemaParser[i]->getTargetNamespace() ==
        type.getNamespace())
            return i;
    return -1;
}


int  WsdlParser::getElementId(Qname & elem)
{
    int i = getSchema(elem);
    if (i >= 0)
        return m_schemaParser[i]->getElementId(elem);

    else
        return -1;
}


int  WsdlParser::getTypeId(Qname & type)
{
    int i = getSchema(type);
    if (i >= 0)
        return m_schemaParser[i]->getTypeId(type);

    else
        return 0;
}


inline string  WsdlParser::getNamespace(string prefix)
{
    return xParser->getNamespace(prefix);
}


void  WsdlParser::error(string s,int level)
{
  if(level==0)
  {
    WsdlException we("Wsdl Parser Exception : " + s);
    we.line = xParser->getLineNumber();
    we.col = xParser->getColumnNumber();
    we.WsdlState = m_State;
    m_ErrorOccured = true;
    throw we;
  }
  else
  {
	   ostr<<"Wsdl parser warning : "<<s<<endl;
  }

}
